# coding: utf-8

# In[1]:


from neuron import h, rxd
import neuron.units
from neuron.units import sec, mM, ms
from neuron.crxd import v
from neuron.crxd.rxdmath import exp, log
from math import pi

h.load_file("stdrun.hoc")


# In[2]:


dir(neuron.units)


# In[3]:


# parameters
h.celsius = 6.3
e = 1.60217662e-19
scale = 1e-14 / e
gnabar = 30e-3 * scale  # molecules/um2 ms mV
gnabar_l = 0.0247 * scale
gkbar = 25e-3 * scale
gkbar_l = 0.05e-3 * scale
gclbar_l = 0.1e-3 * scale
gl = 0.0003 * scale
ukcc2 = 0.3 * mM / sec
unkcc1 = 0.1 * mM / sec
p_max = 0.8 * mM / sec
alpha = 5.3
epsilon = 0.17 * sec * (10**-1)
g_gliamax = 5.0 * mM / sec
e_kmax = 0.25 * sec * (10**-1)
oa_bath = 32.0 / alpha
o_bath = 32.0 * mM
beta_o = 5
avo = rxd.constants.NA()
q10 = 3.0 ** ((h.celsius - 6.3) / 10.0)

# sodium activation 'm'
alpha_m = (0.32 * (v + 54.0)) / (1.0 - exp(-(v + 54.0) / 4.0))
beta_m = (0.28 * (v + 27.0)) / (exp((v + 27.0) / 5.0) - 1.0)
mtau = 1.0 / (q10 * (alpha_m + beta_m))
minf = alpha_m / (alpha_m + beta_m)

# sodium inactivation 'h'
alpha_h = 0.128 * exp(-(v + 50.0) / 18.0)
beta_h = 4.0 / (1.0 + exp(-(v + 27.0) / 5.0))
htau = 1.0 / (q10 * (alpha_h + beta_h))
hinf = alpha_h / (alpha_h + beta_h)

# potassium activation 'n'
alpha_n = (0.032 * (v + 52.0)) / (1.0 - exp(-(v + 52.0) / 5.0))
beta_n = 0.5 * exp(-(v + 57.0) / 40.0)
ntau = 1.0 / (q10 * (alpha_n + beta_n))
ninf = alpha_n / (alpha_n + beta_n)


# In[4]:


soma = h.Section("soma")
soma.pt3dclear()
soma.pt3dadd(-15, 0, 0, 14)
soma.pt3dadd(15, 0, 0, 14)
soma.nseg = 11

glia = h.Section("glia")
glia.pt3dclear()
glia.pt3dadd(-15, 0, 0, 30)
glia.pt3dadd(15, 0, 0, 30)
glia.nseg = 11

vi = 2.0 * pi * (soma.diam / 2.0) ** 2 * soma.L
vo = 100.0 * vi / beta_o

surface_area = 2 * pi * soma.diam / 2.0 * soma.L
volume = 2.0 * pi * (soma.diam / 2) ** 2 * soma.L
glia_surface_area = 2.0 * pi * glia.diam / 2.0 * glia.L
glia_volume = 2.0 * pi * (glia.diam / 2.0) ** 2 * glia.L
volume_scale = avo * volume / surface_area * 10**-18
glia_volume_scale = avo * glia_volume / glia_surface_area * 10**-18


# In[5]:


# initial cli: 6.6; initial clo: 119
# nai = 10 nao = 147
# ki = 125 ko = 2.9


# In[6]:


def concentration(i, o):
    return lambda nd: i if isinstance(nd, rxd.node.Node1D) else o


# In[7]:


# REGIONS------------------------------------------------------------------------------------------------------------------------

# intracellular/extracellular regions
cyt = rxd.Region(soma, name="cyt", nrn_region="i")
mem = rxd.Region(soma, name="cell_mem", geometry=rxd.membrane())
gcyt = rxd.Region(glia, name="cyt", nrn_region="i")
gmem = rxd.Region(glia, name="cell_mem", geometry=rxd.membrane())
dx = vo ** (1.0 / 3.0)
ecs = rxd.Extracellular(-2 * dx, -2 * dx, -2 * dx, 2 * dx, 2 * dx, 2 * dx, dx=dx)
# ecs = rxd.Extracellular(-100, -100, -100, 100, 100, 100, dx=33)
# SPECIES/PARAMETERS------------------------------------------------------------------------------------------------------------------------

# intracellular/extracellular species (Na+, K+, Cl-)
na = rxd.Species(
    [cyt, mem, ecs], name="na", d=1, charge=1, initial=concentration(15.5, 161.5)
)
k = rxd.Species(
    [cyt, mem, ecs], name="k", d=1, charge=1, initial=concentration(142.5, 7.8)
)
cl = rxd.Species(
    [cyt, mem, ecs], name="cl", d=1, charge=-1, initial=concentration(6.0, 130)
)
glial_na = rxd.Parameter(
    [gcyt, gmem], name="na", charge=1, value=18
)  # Na in book is 55 for astrocyte
glial_k = rxd.Parameter([gcyt, gmem], name="k", charge=1, value=80)
dump = rxd.Parameter([cyt, mem, gcyt, gmem], name="dump")
ki, ko, nai, nao, cli, clo, gnai, gki = (
    k[cyt],
    k[ecs],
    na[cyt],
    na[ecs],
    cl[cyt],
    cl[ecs],
    glial_na[gcyt],
    glial_k[gcyt],
)
# i = in cytosol (cyt) o = in extracellular (ecs)
# extracellular oxygen concentration
o2ecs = rxd.Species([ecs], name="o2ecs", initial=oa_bath)

# STATES-------------------------------------------------------------------------------------------------------------------------

# gating variables (m, h, n)
mgate = rxd.State([cyt, mem], name="mgate", initial=0.00787013592322398)
hgate = rxd.State([cyt, mem], name="hgate", initial=0.9981099795551048)
ngate = rxd.State([cyt, mem], name="ngate", initial=0.02284760152971809)

# ALL EQUATIONS------------------------------------------------------------------------------------------------------------------

gna = gnabar * mgate**3 * hgate
gk = gkbar * ngate**4
fko = 1.0 / (1.0 + exp(16.0 - ko))
nkcc1 = unkcc1 * fko * (log((ki * cli) / (ko * clo)) + log((nai * cli) / (nao * clo)))
kcc2 = ukcc2 * log((ki * cli) / (ko * clo))
ena = 26.64 * log(nao / nai)  # nerst equation - reversal potentials
ek = 26.64 * log(ko / ki)  # ^^^^^
ecl = 26.64 * log(cli / clo)  # ^^^
p = p_max / (1.0 + exp((20.0 - oa_bath * alpha) / 3))
pump = (p / (1.0 + exp((25.0 - nai) / 3))) * (1.0 / (1.0 + exp(3.5 - ko)))
gliapump = (
    (1.0 / 3.0) * (p / (1.0 + exp((25.0 - gnai) / 3.0))) * (1.0 / (1.0 + exp(3.5 - ko)))
)
g_glia = g_gliamax / (1.0 + exp(-oa_bath * alpha - 2.5) / 0.2)
glia12 = g_glia / 1.0 + exp((18.0 - ko) / 2.5)

# RATES--------------------------------------------------------------------------------------------------------------------------

# m_gate = rxd.Rate(mgate, (minf - mgate)/mtau)
# h_gate = rxd.Rate(hgate, (hinf - hgate)/htau)
# n_gate = rxd.Rate(ngate, (ninf - ngate)/ntau)

# dm/dt
m_gate = rxd.Rate(mgate, (alpha_m * (1.0 - mgate)) - (beta_m * mgate))

# dh/dt
h_gate = rxd.Rate(hgate, (alpha_h * (1.0 - hgate)) - (beta_h * hgate))

# dn/dt
n_gate = rxd.Rate(ngate, (alpha_n * (1.0 - ngate)) - (beta_n * ngate))

# d[O2]o/dt
o2_ecs = rxd.Rate(o2ecs, (epsilon * (oa_bath - o2ecs)))

# CURRENTS/LEAKS ----------------------------------------------------------------------------------------------------------------

# sodium (Na) current
na_current = rxd.MultiCompartmentReaction(
    nai, nao, gna * (v - ena), mass_action=False, membrane=mem, membrane_flux=True
)

# potassium (K) current
k_current = rxd.MultiCompartmentReaction(
    ki, ko, gk * (v - ek), mass_action=False, membrane=mem, membrane_flux=True
)

# chlorine (Cl) current
cl_current = rxd.MultiCompartmentReaction(
    cli, clo, gclbar_l * (v - ecl), mass_action=False, membrane=mem, membrane_flux=True
)

# nkcc1 (Na+/K+/2Cl- cotransporter)
nkcc1_current = rxd.MultiCompartmentReaction(
    2 * clo + ko + nao,
    2 * cli + ki + nai,
    nkcc1 * volume_scale,
    mass_action=False,
    membrane=mem,
    membrane_flux=True,
)

# kcc2 (K+/Cl- cotransporter)
kcc2_current = rxd.MultiCompartmentReaction(
    clo + ko,
    cli + ki,
    kcc2 * volume_scale,
    membrane=mem,
    mass_action=False,
    membrane_flux=True,
)

# sodium leak
na_leak = rxd.MultiCompartmentReaction(
    nai, nao, gnabar_l * (v - ena), mass_action=False, membrane=mem, membrane_flux=True
)

# potassium leak
k_leak = rxd.MultiCompartmentReaction(
    ki, ko, gkbar_l * (v - ek), mass_action=False, membrane=mem, membrane_flux=True
)

# Na+/K+ pump current in neuron (2K+ in, 3Na+ out)
pump_current = rxd.MultiCompartmentReaction(
    ki,
    ko,
    -2.0 * pump * volume_scale,
    mass_action=False,
    membrane=mem,
    membrane_flux=True,
)
pump_current_other_half = rxd.MultiCompartmentReaction(
    nai,
    nao,
    3.0 * pump * volume_scale,
    mass_action=False,
    membrane=mem,
    membrane_flux=True,
)
oxygen = rxd.MultiCompartmentReaction(
    o2ecs[ecs], dump[cyt], pump * volume_scale, mass_action=False, membrane=mem
)
goxygen = rxd.MultiCompartmentReaction(
    o2ecs[ecs],
    dump[gcyt],
    gliapump * glia_volume_scale,
    mass_action=False,
    membrane=gmem,
)

# Na+/K+ pump current in glia (2K+ in, 3Na+ out)
gliapump_current = rxd.MultiCompartmentReaction(
    2 * gki,
    2 * ko,
    -gliapump * glia_volume_scale,
    mass_action=False,
    membrane=gmem,
    membrane_flux=True,
)
gliapump_current_other_half = rxd.MultiCompartmentReaction(
    3 * gnai,
    3 * nao,
    gliapump * glia_volume_scale,
    mass_action=False,
    membrane=gmem,
    membrane_flux=True,
)


# In[8]:


stimA = h.IClamp(soma(0.5))
stimA.delay = 20
stimA.amp = 1
stimA.dur = 0.2

t = h.Vector().record(h._ref_t)
soma_v = h.Vector().record(soma(0.5)._ref_v)
soma_nai = h.Vector().record(soma(0.5)._ref_nai)
soma_ki = h.Vector().record(soma(0.5)._ref_ki)
soma_cli = h.Vector().record(soma(0.5)._ref_cli)

soma_nao = h.Vector().record(soma(0.5)._ref_nao)
soma_ko = h.Vector().record(soma(0.5)._ref_ko)
soma_clo = h.Vector().record(soma(0.5)._ref_clo)


h.CVode().active(True)
h.finitialize(-70)
# h.dt = 0.025
# h.continuerun(100)
# h.dt = 1
h.continuerun(100)



from . import ctng
from . import surfaces
from . import triangularMesh
import numpy
from numpy import sqrt, fabs


def surface(source, dx=0.25, internal_membranes=False, n_soma_step=100):
    """
    Generates a triangularized mesh of the surface of a neuron.

    Parameters
    ----------
    source : :func:`list`, ``nrn.SectionList``, or ``nrn.Import3D``
        The geometry to mesh.
    dx : double, optional
        Underlying mesh used to generate the triangles.
    internal_membranes : [``True`` | ``False``], optional
        Set to True to not remove internal membranes.
    n_soma_step : integer, optional
        Number of pieces to slice a soma outline into.

    Returns
    -------
    result : :class:`TriangularMesh`
        The mesh.

    Examples
    --------

    A simple meshing of the entire NEURON morphology.

    >>> tri_mesh = geometry3d.surface(h.allsec()) #doctest: +SKIP

    Importing from Neurolucida with a coarser grid.

    >>> h.load_file('stdlib.hoc')
    1.0
    >>> h.load_file('import3d.hoc')
    1.0
    >>> cell = h.Import3d_Neurolucida3()
    >>> cell.input(filename_dot_asc)
    >>> tri_mesh = geometry3d.surface(cell, dx=0.5)

    Removal of the internal membranes is not necessary if the only
    goal is to plot the surface; here we use :mod:`mayavi.mlab`.

    >>> tri_mesh = geometry3d.surface([sec1, sec2, sec3],
    ...                               internal_membranes=True)
    >>> mlab.triangular_mesh(tri_mesh.x, tri_mesh.y, tri_mesh.z,
    ...                      tri_mesh.faces, color=(1, 0, 0))
    >>> mlab.show()

    .. note::
        The use of Import3D objects is recommended over lists of sections
        because the former preserves the soma outline information while
        the later does not. Up to one soma outline is currently supported.
    """
    objects = ctng.constructive_neuronal_geometry(source, n_soma_step, dx)

    xlo = min(obj.xlo for obj in objects)
    ylo = min(obj.ylo for obj in objects)
    zlo = min(obj.zlo for obj in objects)
    xhi = max(obj.xhi for obj in objects)
    yhi = max(obj.yhi for obj in objects)
    zhi = max(obj.zhi for obj in objects)

    # I'm implicitly taking dx = dy = dz here
    # NOTE: triangulate_surface requires consistent discretization
    xs = numpy.arange(xlo - 3 * dx, xhi + 3 * dx, dx)
    ys = numpy.arange(ylo - 3 * dx, yhi + 3 * dx, dx)
    zs = numpy.arange(zlo - 3 * dx, zhi + 3 * dx, dx)

    return triangularMesh.TriangularMesh(
        surfaces.triangulate_surface(objects, xs, ys, zs, internal_membranes)
    )



from testutils import compare_data, tol


def test_multicompartment_reactions_aligment(neuron_instance):
    """A test for multicompartment reactions where one regions has more
    sections than the other.
    """

    h, rxd, data, save_path = neuron_instance

    # Dendrite and spine
    dend = h.Section(name="dend")
    dend.nseg = 101
    dend.pt3dclear()
    dend.pt3dadd(-10, 0, 0, 1)
    dend.pt3dadd(10, 0, 0, 1)

    spine = h.Section(name="spine")
    spine.nseg = 11
    spine.pt3dclear()
    spine.pt3dadd(0, 0, 0, 0.2)
    spine.pt3dadd(0, 1, 0, 0.2)
    spine.connect(dend(0.5))

    # Where -- define the shells and borders
    N = 5  # number of shells -- must be >= 2

    # Where -- shells and border between them
    shells = []
    border = []
    for i in range(N - 1):
        shells.append(
            rxd.Region(
                [dend],
                name="shell%i" % i,
                geometry=rxd.Shell(float(i) / N, (1.0 + i) / N),
            )
        )
        border.append(
            rxd.Region(
                [dend],
                name="border%i" % i,
                geometry=rxd.FixedPerimeter(2.0 * h.PI * (1.0 + i)),
            )
        )

    shells.append(
        rxd.Region(
            [dend, spine],
            nrn_region="i",
            name="shell%i" % (N - 1),
            geometry=rxd.MultipleGeometry(
                secs=[dend, spine], geos=[rxd.Shell((N - 1.0) / N, 1), rxd.inside]
            ),
        )
    )

    # Who -- calcium with an inhomogeneous initial condition
    Dca = 0.6  # um^2/ms

    # scale factor so the flux (Dca/dr)*Ca has units molecules/um^2/ms (where dr is the thickness of the shell)
    mM_to_mol_per_um = 6.0221409e23 * 1e-18
    ca = rxd.Species(
        shells,
        d=Dca,
        name="ca",
        charge=2,
        initial=lambda nd: 1000e-6 if nd.segment in spine else 100e-6,
    )

    # What -- use reactions to setup diffusion between the shells
    cas = []  # calcium on the shells
    for reg in shells:
        cas.append(ca[reg])

    # create the multi-compartment reactions between the pairs of shells
    diffusions = []
    for i in range(N - 1):
        diffusions.append(
            rxd.MultiCompartmentReaction(
                cas[i],
                cas[i + 1],
                mM_to_mol_per_um * Dca,
                mM_to_mol_per_um * Dca,
                border=border[i],
            )
        )

    h.finitialize(-65)
    h.continuerun(2.5)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol



import pytest


@pytest.fixture
def simple_model(neuron_instance):
    """A simple rxd model with species and extracellular regions."""

    h, rxd, data, save_path = neuron_instance
    dend = h.Section(name="dend")
    dend.diam = 2
    dend.nseg = 5
    dend.L = 5
    ecs = rxd.Extracellular(-10, -10, -10, 10, 10, 10, dx=3)
    cyt = rxd.Region(dend, name="cyt", nrn_region="i")
    k = rxd.Species(
        [cyt, ecs],
        name="k",
        d=1,
        charge=1,
        initial=lambda nd: 140 if nd.region == cyt else 3,
    )
    # test initialization for parameters with and without names
    paramA = rxd.Parameter([ecs], name="paramA", initial=1)
    paramB = rxd.Parameter([ecs], initial=0)
    decay = rxd.Rate(k, -0.1 * k)
    model = (dend, cyt, ecs, k, paramA, paramB, decay)
    yield (neuron_instance, model)


def test_ecs_reinit(simple_model):
    """Test rxd.re_init updates extracellular node values from NEURON values"""

    neuron_instance, model = simple_model
    h, rxd, data, save_path = neuron_instance
    dend, cyt, ecs, k, paramA, paramB, decay = model
    h.finitialize(-65)
    dend(0.2).ko = 0
    rxd.re_init()
    assert k[ecs].nodes((0, 0, 0)).value == [0]
    assert dend(0.2).paramAo == 1


def test_ecs_reinit_cvode(simple_model):
    """Test rxd.re_init updates extracellular node values from NEURON segments
    with CVode"""

    neuron_instance, model = simple_model
    h, rxd, data, save_path = neuron_instance
    dend, cyt, ecs, k, paramA, paramB, decay = model
    h.CVode().active(True)
    h.finitialize(-65)
    dend(0.2).ko = 0
    rxd.re_init()
    assert k[ecs].nodes((0, 0, 0)).value == [0]
    assert dend(0.2).paramAo == 1



from .plotshape import PlotShape


def set_backend(backend):
    """sets the GUI backend.

    Currently supported options:
        jupyter -- to plot within a Jupyter notebook
    """
    if backend != "jupyter":
        raise Exception("Unsupported backend")
    else:
        from .config import options

        options["backend"] = backend



from .xml2nrn import *


def biophysics(self, node):
    # following is worthless. need units for Ra, cm, gmax, e
    self.is_physiological_units_ = True
    val = node.get("units")
    if val == "Physiological Units":
        self.is_physiological_units_ = True


def cell(self, node):
    self.biomechs_ = []
    self.cellname = ""
    val = node.get("name")
    if val:
        self.cellname = str(val)
    if debug:
        print("cell")
        self.prattrs(node)


def cell_end(self, node):
    if debug:
        print("endcell")
        for i in range(len(self.cables_)):
            c = self.cables_[i]
            print("Cable ", i, ": ", c.id_, c.parent_cable_id_, c.name_, c.px_, c.pcnt_)

        for bm in self.biomechs_:
            print(bm)

        for cg in self.cablegroups_:
            print(cg)

        for g2i in self.groupname2index_:
            print("groupname2index: ", g2i, " - ", str(self.groupname2index_[g2i]))


def cells(self, node):
    pass


def neuroml(self, node):
    pass



from . import node, rxdmath, constants
import numpy
import weakref
import itertools
from .rxdException import RxDException
from .species import xyz_by_index

_weakref_ref = weakref.ref

# aliases to avoid repeatedly doing multiple hash-table lookups
_itertools_chain = itertools.chain
_numpy_array = numpy.array


def ref_list_with_mult(obj):
    result = []
    for i, p in zip(list(obj.keys()), list(obj.values())):
        w = _weakref_ref(i)
        result += [w] * p
    return result


def get_scheme_rate1_rate2_regions_custom_dynamics_mass_action(args, kwargs):
    """Parse the arguments to a rxd.Reaction or rxd.MultiCompartmentReaction.

    There are four valid options, two for historical
    compatibility, two for future support (these two are the ones
    described in the help)
    """
    if len(args) == 4:
        # bidirectional reaction
        # writing != instead of <> because Python 3 does not support <>
        scheme = args[0] != args[1]
        rate1 = args[2]
        rate2 = args[3]
    elif len(args) == 3:
        # two possibilities which we can distinguish based on if the
        # first argument is an rxdmath._Reaction:
        # 1. with the new way, this would be reactants, products, and a
        #    forward rate
        # 2. with the old way, this is a bidirectional scheme
        if isinstance(args[0], rxdmath._Reaction):
            scheme = args[0]
            rate1 = args[1]
            rate2 = args[2]
        else:
            scheme = args[0] > args[1]
            rate1 = args[2]
            rate2 = None
    elif len(args) == 2:
        # first argument must be a unidirectional rxdmath._Reaction
        # this is the old way and not included in the help because it
        # does not generalize to bidirectional reactions in Python 3
        # because of the missing <>
        scheme = args[0]
        if not isinstance(scheme, rxdmath._Reaction):
            raise RxDException("%r not a recognized reaction scheme" % scheme)
        rate1 = args[1]
        rate2 = None
    else:
        raise RxDException("Invalid number of arguments to rxd.Reaction")

    # keyword arguments
    # custom_dynamics is discouraged in favor of its antonym mass_action
    # (but internally we use custom_dynamics because of how originally
    # designed)
    regions = kwargs.get("regions")
    custom_dynamics = kwargs.get("custom_dynamics")
    mass_action = kwargs.get("mass_action")
    return scheme, rate1, rate2, regions, custom_dynamics, mass_action


class GeneralizedReaction(object):
    """an abstract class, parent of Rate, Reaction, MultiCompartmentReaction"""

    def __del__(self):
        from .rxd import _unregister_reaction

        _unregister_reaction(self)

    def _setup_membrane_fluxes(self, node_indices, cur_map):
        # TODO: make sure this is redone whenever nseg changes
        if not self._membrane_flux:
            return

        from . import species

        sources = [
            r
            for r in self._sources
            if not isinstance(r(), species.SpeciesOnExtracellular)
        ]
        dests = [
            r
            for r in self._dests
            if not isinstance(r(), species.SpeciesOnExtracellular)
        ]

        sources_ecs = [
            r for r in self._sources if isinstance(r(), species.SpeciesOnExtracellular)
        ]
        dests_ecs = [
            r for r in self._dests if isinstance(r(), species.SpeciesOnExtracellular)
        ]
        # locate the regions containing all species (including the one that changes)
        # if all(sptr() for sptr in sources) and all(dptr() for dptr in dests):
        #    active_regions = [r for r in self._regions if all(sptr().indices(r) for sptr in sources + dests)]
        # else:
        #    active_regions = []
        node_indices_append = node_indices.append
        for r in self._active_regions:
            for sec in r._secs:
                for seg in sec:
                    node_indices_append(seg.node_index())
        self._do_memb_scales(cur_map)

    def _get_args(self, states):
        args = []
        args_append = args.append
        self_indices_dict = self._indices_dict
        for sptr in self._involved_species:
            s = sptr()
            if not s:
                return None
            args_append(states[self_indices_dict[s]])
        return args

    def _update_indices(self):
        # this is called anytime the geometry changes as well as at init
        from . import species

        # Default values
        self._indices_dict = {}
        self._indices = []
        # self._jac_rows = []
        # self._jac_cols = []
        self._mult = [1]
        self._mult_extended = self._mult
        active_secs = None

        # filter regions with no sections
        def spfilter(sp):
            if not sp():
                return False
            regs = (
                sp()._regions if isinstance(sp(), species.Species) else [sp()._region()]
            )
            for r in regs:
                if r and (any(r._secs1d) or any(r._secs3d)):
                    break
            else:
                return False
            return True

        sources = list(
            filter(
                spfilter,
                [
                    r
                    for r in self._sources
                    if not isinstance(r(), species.SpeciesOnExtracellular)
                ],
            )
        )
        dests = list(
            filter(
                spfilter,
                [
                    r
                    for r in self._dests
                    if not isinstance(r(), species.SpeciesOnExtracellular)
                ],
            )
        )
        sources_ecs = [
            r for r in self._sources if isinstance(r(), species.SpeciesOnExtracellular)
        ]
        dests_ecs = [
            r for r in self._dests if isinstance(r(), species.SpeciesOnExtracellular)
        ]

        sp_regions = None
        if self._trans_membrane and (
            sources or dests
        ):  # assume sources share common regions and destinations share common regions
            sp_regions = list(
                {sptr()._region() for sptr in sources}.union(
                    {sptr()._region() for sptr in dests}
                )
            )
        elif sources and dests:
            sp_regions = list(
                set.intersection(
                    *[
                        set(sptr()._regions)
                        if isinstance(sptr(), species.Species)
                        else {sptr()._region()}
                        for sptr in sources + dests
                    ]
                )
            )

        # The reactants do not share a common region
        if not sp_regions:
            active_regions = [
                s()._extracellular()._region for s in sources_ecs + dests_ecs if s()
            ]
            # if a region is specified the reaction should only take place there
            if self._regions != [None]:
                self._active_regions = self._regions
            # alternatively if SpeciesOnExtracellular were specified the
            # reaction should only take place on those extracellular regions
            elif active_regions:
                self._active_regions = active_regions

            if hasattr(self, "_active_regions"):
                for reg in self._active_regions:
                    if not hasattr(reg, "_secs1d") or (
                        any(reg._secs1d) or any(reg._secs3d)
                    ):
                        break
                else:
                    if not sources_ecs or not dests_ecs:
                        return

            # if neither were specified don't set the '_has_regions' attribute
            # so the reaction takes place everywhere the species is defined
            for sptr in self._involved_species:
                self._indices_dict[sptr()] = []
            # Setup for extracellular
            self._mult = list(-1 for v in self._sources) + list(1 for v in self._dests)
            self._mult = _numpy_array(self._mult)
            return

        from .multiCompartmentReaction import MultiCompartmentReaction

        # locate the regions containing all species (including the one that changes)
        # (we do not need every species to be on the same region for multi compartment reactions)
        if all(sptr() for sptr in sources) and all(dptr() for dptr in dests):
            active_regions = [
                r
                for r in self._regions
                if all(sptr().indices(r) for sptr in sources + dests)
            ]
        else:
            active_regions = []
        # MultiCompartmentReactions do not require involved species to share common regions
        if not isinstance(self, MultiCompartmentReaction):
            for sptr in self._involved_species:
                s = sptr()
                if s and not isinstance(s, species.SpeciesOnExtracellular):
                    for r in self._regions:
                        if r in active_regions and not s.indices(r):
                            del active_regions[active_regions.index(r)]
                elif s and isinstance(s, species.SpeciesOnExtracellular):
                    r = s._extracellular()._region
                    if r in active_regions:
                        del active_regions[active_regions.index(r)]
                else:
                    active_regions = []

        def intersection(los):
            if los:
                return set.intersection(*los)
            return None

        # If we haven't identified active_regions -- use the regions where all species are defined
        if len(active_regions) == 0 or active_regions == [None]:
            if self._trans_membrane:
                src_regions = intersection(
                    [
                        set(sptr()._regions)
                        if isinstance(sptr(), species.Species)
                        else {sptr()._region()}
                        for sptr in sources
                    ]
                )
                if not src_regions:
                    raise RxDException(
                        "Error in %r. The source species do not share a common region"
                        % self
                    )
                src_sections = intersection(
                    [set(reg.secs) for reg in src_regions if reg is not None]
                )
                dest_regions = intersection(
                    [
                        set(sptr()._regions)
                        if isinstance(sptr(), species.Species)
                        else {sptr()._region()}
                        for sptr in dests
                    ]
                )
                if not dest_regions:
                    raise RxDException(
                        "Error in %r. The destination species do not share a common region"
                        % self
                    )
                dest_sections = intersection(
                    [set(reg.secs) for reg in dest_regions if reg is not None]
                )
                active_regions = set.union(src_regions, dest_regions)
                active_secs = set.union(src_sections, dest_sections)
            else:
                active_regions = list(
                    intersection(
                        [
                            set(sptr()._regions)
                            if isinstance(sptr(), species.Species)
                            else {sptr()._region()}
                            for sptr in sources + dests
                        ]
                    )
                )
                if not active_regions:
                    raise RxDException(
                        "Error in %r. The species do not share a common region" % self
                    )
                active_secs = set.intersection(
                    *[set(reg.secs) for reg in active_regions if reg]
                )
        else:
            active_secs = set.intersection(
                *[set(reg.secs) for reg in active_regions if reg]
            )

        self._active_regions = active_regions

        if isinstance(self, MultiCompartmentReaction):
            sources = [
                r
                for r in self._sources
                if not isinstance(r(), species.SpeciesOnExtracellular)
            ]
            dests = [
                r
                for r in self._dests
                if not isinstance(r(), species.SpeciesOnExtracellular)
            ]

            # flux occurs on sections which have both source, destination and membrane
            active_secs_list = list(self._regions[0]._secs1d)
            for sp in sources + dests:
                if sp() and sp()._region():
                    active_secs_list = [
                        sec for sec in active_secs_list if sec in sp()._region()._secs1d
                    ]
        else:
            active_secs_list = [
                sec
                for reg in active_regions
                if reg
                for sec in reg.secs
                if sec in active_secs
            ]

        # store the indices
        for sptr in self._involved_species:
            s = sptr()
            if not isinstance(s, species.SpeciesOnExtracellular):
                self._indices_dict[s] = s.indices(active_regions, active_secs)
        sources_indices = [
            sptr().indices(active_regions, active_secs) for sptr in sources
        ]
        dests_indices = [dptr().indices(active_regions, active_secs) for dptr in dests]
        self._indices = sources_indices + dests_indices
        volumes, surface_area, diffs = node._get_data()
        # self._mult = [list(-1. / volumes[sources_indices]) + list(1. / volumes[dests_indices])]
        if self._trans_membrane and active_regions:
            # note that this assumes (as is currently enforced) that if trans-membrane then only one region

            molecules_per_mM_um3 = constants.molecules_per_mM_um3()

            # TODO: verify the areas and volumes are in the same order!
            areas = _numpy_array(
                list(
                    _itertools_chain.from_iterable(
                        [
                            list(self._regions[0]._geometry.volumes1d(sec))
                            for sec in active_secs_list
                        ]
                    )
                )
            )
            if not self._scale_by_area:
                areas = numpy.ones(len(areas))
            if len(sources_ecs) == len(dests_ecs) == 0:
                self._mult = [
                    -areas / volumes[si] / molecules_per_mM_um3
                    for si in sources_indices
                ] + [areas / volumes[di] / molecules_per_mM_um3 for di in dests_indices]
            # TODO: check for multicompartment reaction within the ECS
            elif len(sources_ecs) > 0 and len(dests_ecs) == 0:
                self._mult = [
                    -areas
                    / (
                        numpy.prod(s()._extracellular()._dx)
                        * s().alpha_by_location(xyz_by_index(di))
                    )
                    / molecules_per_mM_um3
                    for s, di in zip(sources_ecs, dests_indices)
                ] + [areas / volumes[di] / molecules_per_mM_um3 for di in dests_indices]
            elif len(sources_ecs) == 0 and len(dests_ecs) > 0:
                self._mult = [
                    -areas / volumes[si] / molecules_per_mM_um3
                    for si in sources_indices
                ] + [
                    areas
                    / (
                        numpy.prod(s()._extracellular()._dx)
                        * s().alpha_by_location(xyz_by_index(si))
                    )
                    / molecules_per_mM_um3
                    for s, si in zip(dests_ecs, sources_indices)
                ]

            else:
                # If both the source & destination are in the ECS, they should use a reaction
                # not a multicompartment reaction
                RxDException(
                    "An extracellular source and destination is not possible with a multi-compartment reaction."
                )
        else:
            self._mult = list(-1 for v in sources_indices) + list(
                1 for v in dests_indices
            )
        self._mult = _numpy_array(self._mult)
        self._update_jac_cache()

    def _evaluate(self, states):
        """returns: (list of lists (lol) of increase indices, lol of decr indices, list of changes)"""
        args = self._get_args(states)
        if args is None:
            return ([], [], [])
        return self._evaluate_args(args)

    def _evaluate_args(self, args):
        return (self._indices, self._mult, self._rate(*args))

    def _get_memb_flux(self, states):
        if self._membrane_flux:
            # TODO: refactor the inside of _evaluate so can construct args in a separate function and just get self._rate() result
            rates = self._evaluate(states)[2]
            return self._memb_scales * rates
        else:
            return []

    def _update_jac_cache(self):
        from . import species

        num_involved = len(self._involved_species)
        # self._jac_rows = list(_itertools_chain(*[ind * num_involved for ind in self._indices]))
        num_ind = len(self._indices)
        # self._jac_cols = list(_itertools_chain(*[self._indices_dict[s()] for s in self._involved_species if not isinstance(s(),species.SpeciesOnExtracellular)])) * num_ind
        if self._trans_membrane:
            self._mult_extended = self._mult
            # self._mult_extended = [sum([list(mul) * num_involved], []) for mul in self._mult]
            # self._mult_extended = [list(_itertools_chain.from_iterable(list(mul) * num_involved)) for mul in self._mult]
        else:
            self._mult_extended = self._mult


"""
    def _jacobian_entries(self, states, multiply=1, dx=1.e-10):
        args = self._get_args(states)
        indices, mult, base_value = self._evaluate_args(args)
        mult = self._mult_extended
        derivs = []
        derivs_append = derivs.append
        for i, arg in enumerate(args):
            args[i] = arg + dx
            new_value = self._evaluate_args(args)[2]
            args[i] = arg
            derivs_append((new_value - base_value) / dx)
        derivs = _numpy_array(list(_itertools_chain(*derivs)))
        if self._trans_membrane:
            data = list(_itertools_chain(*[derivs * mul * multiply for mul in mult]))
            #data = derivs * mult * multiply
        else:
            data = list(_itertools_chain(*[derivs * mul * multiply for mul in mult]))
        return self._jac_rows, self._jac_cols, data
"""



import neuron
import numpy


class ScalarField:
    """
    A scalar field.
    """

    def __init__(self, xlo, xhi, ylo, yhi, zlo, zhi, dx, dtype=None):
        """
        Parameters
        ----------
        xlo : double
            Minimum x value.
        xhi : double
            Maximum x value.
        ylo : double
            Minimum y value.
        yhi : double
            Maximum y value.
        zlo : double
            Minimum z value.
        zhi : double
            Maximum z value.
        dx : double
            Spatial step size.
        dtype : string, optional
            Passed to :mod:`numpy` to define the data type.

        .. note
            The ScalarField may contain locations slightly beyond `(xhi, yhi, zhi)`
            if the extrema are not evenly divisible by `dx`, however the center
            point of the voxels will not exceed the bounds.
        """
        # I'm implicitly taking dx = dy = dz here
        xs = numpy.arange(xlo, xhi, dx)
        ys = numpy.arange(ylo, yhi, dx)
        zs = numpy.arange(zlo, zhi, dx)
        # shift everything by half a voxel
        xs = xs + dx * 0.5
        ys = ys + dx * 0.5
        zs = zs + dx * 0.5
        if xs[-1] > xhi:
            xs = xs[:-1]
        if ys[-1] > yhi:
            ys = ys[:-1]
        if zs[-1] > zhi:
            zs = zs[:-1]
        self._xs = xs
        self._ys = ys
        self._zs = zs
        self._dx = dx
        self._values = numpy.zeros(
            [len(self._xs), len(self._ys), len(self._zs)], dtype=dtype
        )

    @property
    def shape(self):
        """The shape of the scalar field.

        This is a convenience property and is equal to self._values.shape."""
        return self._values.shape

    def __getitem__(self, *args):
        """
        Indexing in to the values of the mesh. This is a syntactic shorthand for `values[indices]`.

        Examples
        --------

        To return the value from `mesh` whose position is `mesh.xs[0]`, `mesh.ys[1]`, `mesh.zs[2]`:

        >>> a = mesh.__getitem__(0, 1, 2)

        Equivalently, one can use the syntactic shorthand []:

        >>> a = mesh[0, 1, 2]

        Slices are also supported:

        >>> a = mesh[0 : 2, :, 1]
        """
        return self._values(*args)

    @property
    def dx(self):
        """Discretization size of the grid."""
        return self._dx

    @property
    def xs(self):
        """The x coordinates of the centers of the voxels."""
        return self._xs

    @property
    def ys(self):
        """The y coordinates of the centers of the voxels."""
        return self._ys

    @property
    def zs(self):
        """The z coordinates of the centers of the voxels."""
        return self._zs

    @property
    def values(self):
        """State values."""
        return self._values



# Error tests for netpar.cpp
# Some minor coverage increase for netpar.cpp when nhost > 1

# Prepend the location of the current script to the search path, so we can
# import from test_hoc_po.
import os, sys

sys.path.insert(0, os.path.dirname(os.path.realpath(__file__)))

from neuron import h

pc = h.ParallelContext()

from neuron.expect_hocerr import expect_err
from test_hoc_po import Ring

cvode = h.CVode()


def run(tstop):
    pc.set_maxstep(1)
    h.finitialize(-65)
    pc.psolve(tstop)


def mpi_test1():  # pgvts_deliver and pc.spike_record
    dtsav = h.dt
    r = Ring(3, 2)
    for gid, cell in r.cells.items():
        if type(cell) == type(h):
            cell.axon.disconnect()
            pc.multisplit(cell.axon(0), gid)
            pc.multisplit(cell.soma(1), gid)

    pc.multisplit()
    gids = h.Vector([gid for gid in r.cells])
    pc.spike_record(gids, r.spiketime, r.spikegid)
    cvode.active(1)
    cvode.condition_order(1)  # investigate why condition_order(2) fails
    cvode.debug_event(1)
    run(3.0)

    cvode.queue_mode(1, 0)
    pc.spike_compress(3, 1)

    cvode.active(0)
    h.dt = dtsav
    pc.gid_clear()
    cvode.debug_event(0)
    cvode.condition_order(1)


def mpi_test2():
    return


def err_test1():
    r = Ring(3, 2)
    expect_err("pc.set_gid2node(0, pc.id())")
    nc = pc.gid_connect(10000, r.cells[0].syn)
    expect_err("pc.set_gid2node(10000, pc.id())")
    expect_err("pc.threshold(10000)")
    expect_err("pc.cell(10000)")
    expect_err("pc.cell(99999)")
    expect_err("pc.cell(0, None)")
    expect_err("pc.gid_connect(0, None)")
    pc.set_gid2node(99999, pc.id())
    expect_err("pc.gid_connect(99999, r.cells[0].syn)")
    expect_err("pc.gid_connect(1, r.cells[0].syn, h.Vector())")
    nc = h.NetCon(None, r.cells[1].syn)
    expect_err("pc.gid_connect(1, r.cells[0].syn, nc)")
    nc = h.NetCon(None, r.cells[0].syn)
    pc.gid_connect(1, r.cells[0].syn, nc)

    pc.gid_clear()
    del nc, r
    locals()


def err_test2():
    r = Ring(3, 2)
    r.ncs[0].delay = 0
    pc.nthread(2)
    pc.set_maxstep(1)
    expect_err("h.finitialize(-65)")
    pc.nthread(1)
    r.ncs[0].delay = h.dt
    pc.set_maxstep(h.dt)
    h.finitialize(-65)
    if "'NRN_ENABLE_MPI=OFF'" not in h.nrnversion(6):
        # Fixed step method does not allow a mindelay < dt + 1e-10
        expect_err("pc.psolve(1.0)")  # wonder if this is too stringent an error
    else:
        pc.psolve(1.0)
    r.ncs[0].delay = 0
    assert pc.set_maxstep(1) == 1.0
    cvode.queue_mode(0, 1)
    pc.set_maxstep(1)
    h.finitialize(-65)
    pc.psolve(1)
    assert cvode.queue_mode() == 0
    v1 = h.Vector(10)
    pc.max_histogram(v1)
    v2 = h.Vector(10)
    pc.max_histogram()
    # cover nrn_gid2outpresyn
    pc.prcellstate(10000, "tmp")
    pc.prcellstate(0, "tmp")

    pc.gid_clear()
    del r
    locals()


def test_1():
    if pc.nhost() > 1:
        mpi_test1()
        mpi_test2()
    else:
        err_test1()
        err_test2()


if __name__ == "__main__":
    test_1()
    pc.barrier()
    h.quit()



from neuron import h, rxd
from neuron.rxd import v
from neuron.rxd.rxdmath import exp, log
import numpy

h.load_file("stdrun.hoc")

# set the number of rxd threads
rxd.nthread(2)

# set the seed
numpy.random.seed(63245)

# enable extracellular rxd (only necessary for NEURON prior to 7.7)
rxd.options.enable.extracellular = True

# Model parameters
e = 1.60217662e-19
scale = 1e-14 / e
h.celsius = 37

# Persistent Na current parameters
gnap = 0.4e-3 * scale
thmp = -40
thhp = -48
sigmp = 6
sighp = -6
nap_minf = 1.0 / (1.0 + exp(-(v - thmp) / sigmp))
nap_hinf = 1.0 / (1.0 + exp(-(v - thhp) / sighp))

# Fast Na current parameters
gna = 3e-3 * scale
thm = -34
sigm = 5

# K current parameters
gk = 5e-3 * scale
thn = -55.0
sgn = 14.0
taun0 = 0.05
taun1 = 0.27
thnt = -40
sn = -12
phin = 0.8
minf = 1.0 / (1.0 + exp(-(v - thm) / sigm))
ninf = 1.0 / (1.0 + exp(-(v - thn) / sgn))
taun = taun0 + taun1 / (1 + exp(-(v - thnt) / sn))

# K-leak parameters
gl = 0.1e-3 * scale
el = -70

# leak parameters
pas_gl = 0.3
pas_el = -70

# Simulation parameters
Lx, Ly, Lz = 500, 500, 100  # size of the extracellular space mu m^3
Ncell = 2  # number of neurons
somaR = 40

# Extracellular rxd
# Where? -- define the extracellular space
ecs = rxd.Extracellular(
    -Lx / 2.0,
    -Ly / 2.0,
    -Lz / 2.0,
    Lx / 2.0,
    Ly / 2.0,
    Lz / 2.0,
    dx=10,
    volume_fraction=0.2,
    tortuosity=1.6,
)

# Who? -- define the species
k = rxd.Species(
    ecs,
    name="k",
    d=2.62,
    charge=1,
    initial=lambda nd: 40 if nd.x3d**2 + nd.y3d**2 + nd.z3d**2 < 50**2 else 3.5,
)

na = rxd.Species(ecs, name="na", d=1.78, charge=1, initial=134)
ko, nao = k[ecs], na[ecs]

# What?
# No extracellular reactions, just diffusion.


class Neuron:
    """A neuron with soma and fast and persistent sodium
    currents, potassium currents, passive leak and potassium leak and an
    accumulation mechanism."""

    def __init__(self, x, y, z, rec=False):
        self.x = x
        self.y = y
        self.z = z

        self.soma = h.Section(name="soma", cell=self)
        # add 3D points to locate the neuron in the ECS
        self.soma.pt3dadd(x, y, z + somaR, 2.0 * somaR)
        self.soma.pt3dadd(x, y, z - somaR, 2.0 * somaR)

        # Where? -- define the intracellular space and membrane
        self.cyt = rxd.Region(self.soma, name="cyt", nrn_region="i")
        self.mem = rxd.Region(self.soma, name="mem", geometry=rxd.membrane())
        cell = [self.cyt, self.mem]

        # Who? -- the relevant ions and gates
        self.k = rxd.Species(cell, name="k", d=2.62, charge=1, initial=125)
        self.na = rxd.Species(cell, name="na", d=1.78, charge=1, initial=10)
        self.n = rxd.State(cell, name="n", initial=0.25512)
        self.ki, self.nai = self.k[self.cyt], self.na[self.cyt]

        # What? -- gating variables and ion currents
        self.n_gate = rxd.Rate(self.n, phin * (ninf - self.n) / taun)

        # Nernst potentials
        ena = 1e3 * h.R * (h.celsius + 273.15) * log(nao / self.nai) / h.FARADAY
        ek = 1e3 * h.R * (h.celsius + 273.15) * log(ko / self.ki) / h.FARADAY

        # Persistent Na current
        self.nap_current = rxd.MultiCompartmentReaction(
            self.nai,
            nao,
            gnap * nap_minf * nap_hinf * (v - ena),
            mass_action=False,
            membrane=self.mem,
            membrane_flux=True,
        )
        # Na current
        self.na_current = rxd.MultiCompartmentReaction(
            self.nai,
            nao,
            gna * minf**3 * (1.0 - self.n) * (v - ena),
            mass_action=False,
            membrane=self.mem,
            membrane_flux=True,
        )
        # K current
        self.k_current = rxd.MultiCompartmentReaction(
            self.ki,
            ko,
            gk * self.n**4 * (v - ek),
            mass_action=False,
            membrane=self.mem,
            membrane_flux=True,
        )
        # K leak
        self.k_leak = rxd.MultiCompartmentReaction(
            self.ki,
            ko,
            gl * (v - ek),
            mass_action=False,
            membrane=self.mem,
            membrane_flux=True,
        )
        # passive leak
        self.soma.insert("pas")
        self.soma(0.5).pas.g = pas_gl
        self.soma(0.5).pas.e = pas_el

        if rec:  # record membrane potential (shown in figure 1C)
            self.somaV = h.Vector()
            self.somaV.record(self.soma(0.5)._ref_v, rec)


# Randomly distribute the neurons
all_neurons = [
    Neuron(
        (numpy.random.random() * 2.0 - 1.0) * (Lx / 2.0),
        (numpy.random.random() * 2.0 - 1.0) * (Ly / 2.0),
        (numpy.random.random() * 2.0 - 1.0) * (Lz / 2.0),
        100,
    )
    for i in range(0, Ncell)
]


# initialize and set the intracellular concentrations
h.finitialize(-70)


h.dt = 1  # use a large time step as we are not focusing on spiking behaviour
# but on slower diffusion

h.continuerun(100)

if __name__ == "__main__":
    from matplotlib import pyplot

    pyplot.imshow(k[ecs].states3d.mean(2), extent=k[ecs].extent("xy"))
    for sec in h.allsec():
        pyplot.plot(
            [sec.x3d(i) for i in range(sec.n3d())],
            [sec.y3d(i) for i in range(sec.n3d())],
            "o",
        )

    pyplot.colorbar()



"""plugins.py

Functions for allowing solver plugins/overrides.
"""
from .rxdException import RxDException
from . import rxd
from . import options
from . import species
from . import node
from . import section1d


class SolverPlugin:
    def __init__(self):
        _initialize(self)

    """Interface for solver plugins."""

    def init(self, species_list, reactions_list):
        """Perform any needed initialization.

        This function may be called multiple times. Subsequent calls purge all
        data and restart the solver de novo.
        """
        self._species = []
        self._reactions = []
        self._values = []
        # NOTE: to use shared memory, simply set
        #           self._values = neuron.rxd.node._states
        #       and define self._index appropriately
        #
        # NOTE: when not using shared memory, self._values should be an object
        #       that transfers data only on demand

        self._set_species(species_list)
        self._set_reactions(reactions_list)
        self._set_values()

    def _set_values(self):
        """update self._values if necessary"""

    def _set_species(self, species_list):
        """Set the species to be used.

        They should be stored internally with weakrefs.

        The plugin will likely need this information for at least five reasons:
            1. knowing how many species there are
            2. being able to identify the indices corresponding to a given node
            3. grabbing the mesh
            4. transferring states to/from the solver
            5. diffusion rates

        This function raises an RxDException if the plugin does not support
        the species options.
        """
        import weakref

        # TODO: use a callback to automatically invalidate the plugin state
        #       if a species is destroyed
        self._species = [weakref.proxy(s) for s in species_list]
        self._set_mesh()

    def _set_reactions(self, reactions_list):
        """Store the reactions.

        Invokes self._process_reactions to do any necessary preprocessing.
        """
        import weakref

        # TODO: use a callback to automatically invalidate the plugin state
        #       if a reaction is destroyed
        self._reactions = [weakref.proxy(s) for s in reactions_list]
        self._process_reactions()

    def _process_reactions(self):
        """Do any necessary preprocessing of the reactions."""

    def _set_mesh(self):
        """Use self._species to define a mesh."""

    def _index(self, node):
        """Return the index of a node in the plugin's internal storage."""

    def value(self, node):
        """Return the value (concentration or mass) associated with the node."""
        return self._values_from_indices([self._index(node)])

    def _values_from_indices(self, indices):
        """Return a vector of values associated with a vector of indices."""
        import numpy

        return numpy.asarray(self._values)[indices]

    def advance(self, dt):
        """Advance the solver by dt.

        By default, the solver will use its local copy of state variables.
        This behavior can be overriden by explicitly calling
        transfer_states_to_solver at the beginning of the advance function and
        calling transfer_states_to_neuron at the end. These two function calls
        allow pointers to concentrations to work at the expense of increased
        run-time.
        """

    def supported_sim(self):
        """Returns True if this simulation is supported; else False"""
        return False

    def set_rate_of_change(self, nodes, rates_of_change):
        """Sets steady influx to nodes."""
        import numpy

        self._roc = numpy.zeros(len(self._values))
        for node, r in zip(nodes, rates_of_change):
            self._roc[self._index(node)] += r

    def transfer_states_from_neuron(self):
        """Transfers all state variables to the solver."""
        # NOTE: the default approach described below is general but consequently
        #       inefficient; plugin developers should override this function
        for species in self._species:
            try:
                for node in species.nodes:
                    self._values[self._index(node)] = node.value
            except ReferenceError:
                pass

    def transfer_states_to_neuron(self, which=None):
        """Transfers state variables from the solver to NEURON.

        If which is None, transfer all states; otherwise, transfer at least
        the requested states.

        If the type of which is not supported, transfer all states.
        """

        # NOTE: the default approach described below is general but consequently
        #       inefficient; plugin developers should override this function
        for species in self._species:
            try:
                for node in species.nodes:
                    node.value = self._values[self._index(node)]
            except ReferenceError:
                pass


class SharedMemorySolverPlugin(SolverPlugin):
    def _set_values(self):
        self._values = node._states

    def transfer_states_from_neuron(self):
        """no need to transfer states since shared memory"""

    def transfer_states_to_neuron(self):
        """no need to transfer states since shared memory"""

    def _index(self, node):
        """shared memory, so trivial index mapping"""
        return node._index


def store_simple_3d_mesh(plugin):
    mesh = None
    for s in plugin._species:
        if s._dimension != 3:
            raise RxDException("This plugin only supports 3D simulations")
        if len(s._regions) != 1:
            raise RxDException("This plugin only supports 1 region")
        if mesh is None:
            mesh = s._regions[0]._mesh
        if mesh != s._regions[0]._mesh:
            raise RxDException("This plugin requires a unique mesh")

    plugin._mesh = mesh._values
    plugin._dx = mesh._dx


def _cvode_error(*args, **kwargs):
    raise RxDException(
        "Variable step method not supported for external plugin solvers."
    )


def set_solver(plugin_solver_object=None):
    """Set or remove a plugin solver for reaction-diffusion dynamics.

    If plugin_solver_object is None, removes any plugin solver.

    Otherwise:

    The plugin_solver_object must support the rxd.SolverPlugin interface,
    although it need subclass it.

    NOTE: External solver support currently requires the model to be simulated
          in three dimensions. 1D branching behavior is undefined.

    NOTE: For now, only fixed step advances are supported.

    NOTE: Correct pointer access to read/write state variables is not guaranteed
          in general, but can be obtained by having the advance method sync
          state values from neuron at the beginning and to neuron at the end.
    """
    if plugin_solver_object is None:
        _default_solver()
    else:
        change_currents = lambda rhs: None
        change_conductance = None
        rxd._setup = rxd._orig_setup
        rxd._currents = change_currents
        rxd._ode_count = _cvode_error
        rxd._ode_reinit = _invalidate_plugin
        rxd._ode_fun = _cvode_error
        rxd._ode_solve = _cvode_error
        rxd._fixed_step_solve = lambda dt: _fixed_step_advance(dt, plugin_solver_object)
        rxd._ode_jacobian = _cvode_error

        rxd._external_solver = plugin_solver_object
    _invalidate_plugin()


def _invalidate_plugin():
    rxd._external_solver_initialized = False


def _initialize(solver):
    all_species = [s() for s in species._get_all_species() if s() is not None]
    all_reactions = [r() for r in rxd._all_reactions if r() is not None]

    solver.init(all_species, all_reactions)

    # TODO: transfer information about the reactions

    if not solver.supported_sim():
        raise RxDException(
            "Selected plugin solver does not support the current simulation."
        )

    solver.transfer_states_from_neuron()
    rxd._external_solver_initialized = True


def _fixed_step_advance(dt, solver):
    if not rxd._external_solver_initialized:
        _initialize(solver)

    import warnings

    warnings.warn("Plugin support for membrane currents not yet implemented.")

    solver.set_rate_of_change([], [])

    # do the advance
    solver.advance(dt)

    # TODO: only collect values that are must-sync
    solver.transfer_states_to_neuron()

    # transfer any 1D states to the legacy grid
    section1d._transfer_to_legacy()

    # TODO: transfer any 3D states to the legacy grid


def _default_solver():
    """Restore the default solver."""
    rxd._setup = rxd._orig_setup
    rxd._currents = rxd._orig_currents
    rxd._ode_count = rxd._orig_ode_count
    rxd._ode_reinit = rxd._orig_ode_reinit
    rxd._ode_fun = rxd._orig_ode_fun
    rxd._ode_solve = rxd._orig_ode_solve
    rxd._fixed_step_solve = rxd._orig_fixed_step_solve
    rxd._ode_jacobian = _orig_ode_jacobian

    rxd._external_solver = None
    rxd._external_solver_initialized = False



#!/usr/bin/env python3
"""Script to compare output of [Core]NEURON runs.

This utility is used to compare the results of different NEURON and CoreNEURON
runs and determine whether or not they are consistent with one another.
Positional arguments are of the form testname::type1::path1[::type2::path2]...

  Typical usage example:

  compare_test_results.py test1::asciispikes::path1 test2::asciispikes::path2
"""
import collections
import difflib
import glob
import itertools
import string
import sys


def parse_arguments(argv):
    parsed_data = collections.defaultdict(dict)

    def pairwise(iterable):
        "pairwise('ABCD') --> AB CD"
        return itertools.zip_longest(*[iter(iterable)] * 2)

    for test_spec in argv:
        test_spec_components = test_spec.split("::")
        if len(test_spec_components) % 2 != 1:
            raise Exception("Malformed test specification: " + test_spec)
        test_name = test_spec_components[0]
        for data_type, data_path in pairwise(test_spec_components[1:]):
            if test_name in parsed_data[data_type]:
                raise Exception(
                    "Got duplicate entry for data type {} in test {}".format(
                        data_type, test_name
                    )
                )
            parsed_data[data_type][test_name] = data_path
    return parsed_data


def load_ascii_spikes(data_pattern):
    spike_data = []
    for data_path in sorted(glob.glob(data_pattern)):
        with open(data_path, "r") as data_file:
            for line in data_file:
                time, gid = line.strip().split(maxsplit=1)
                spike_data.append((float(time), int(gid)))
    return spike_data


def check_compatibility(data_type, test_data):
    # TODO: support more data types, compare HDF5 files etc.
    if data_type != "asciispikes":
        raise Exception("Only {} test output is supported".format(data_type))
    # Load the ASCII output files
    sorted_data = []
    for test_name, data_path in test_data.items():
        spike_data = load_ascii_spikes(data_path)
        spike_data.sort()  # sort by time, then by ID
        sorted_data.append((test_name, spike_data))
    # Clearly this is a bit excessive because the relationship is transitive,
    # but hopefully the output will be helpful when finding the source of a
    # failure.
    results = []
    names = sorted([name for name, _ in sorted_data])
    matrix = [[None] * len(names) for _ in range(len(names))]
    for (name1, data1), (name2, data2) in itertools.combinations(sorted_data, 2):
        result = data1 == data2  # TODO fuzzier comparison?
        name1_index, name2_index = names.index(name1), names.index(name2)
        min_index, max_index = (
            min(name1_index, name2_index),
            max(name1_index, name2_index),
        )
        matrix[min_index][max_index] = result
        print(" ".join([name1, "matches" if result else "DOES NOT match", name2]))
        if not result:
            # Try and print a helpful diff, this is a bit inelegant but it uses standard library...
            def as_strings(data):
                return [str(x) for x in data]

            for line in difflib.unified_diff(
                as_strings(data1),
                as_strings(data2),
                fromfile=name1,
                tofile=name2,
                lineterm="",
            ):
                print(line)
        results.append(result)
    print(" ".join(string.ascii_uppercase[i] for i in range(len(names))))
    print("-" * 2 * len(names))

    def val(row, col):
        if col < row:
            return " "
        return {None: ".", True: "M", False: "x"}[matrix[row][col]]

    for i in range(len(names)):
        print(
            " ".join(val(i, j) for j in range(len(names)))
            + " | "
            + string.ascii_uppercase[i]
            + " = "
            + names[i]
        )
    if any(results):
        print("M = match")
    if not all(results):
        print("x = mismatch")
    return all(results)


if __name__ == "__main__":
    try:
        test_specs = parse_arguments(sys.argv)
    except Exception as e:
        print(__doc__)
        print("ERROR parsing arguments: " + str(e))
        sys.exit(1)
    results = [
        check_compatibility(data_type=data_type, test_data=test_data)
        for data_type, test_data in test_specs.items()
    ]
    sys.exit(not all(results))



import ctypes
import numpy
from .. import options
from neuron import nrn_dll_sym
from ctypes import c_double, POINTER


"""Function to take any surface vox with information on distance of each vertex from surface 
and output approx. surface area contained in that vox"""


def surface_area(itemlist, vox, grid):
    res = options.ics_partial_surface_resolution
    i0, j0, k0 = vox
    rng = range(res + 1)
    rng0 = range(res)
    dx, dy, dz = grid["dx"], grid["dy"], grid["dz"]
    Sx, Sy, Sz = dx / res, dy / res, dz / res
    x0 = grid["xlo"] + i0 * dx
    y0 = grid["ylo"] + j0 * dy
    z0 = grid["zlo"] + k0 * dz
    dist = {}
    for i in rng:
        x = x0 + i * Sx
        for j in rng:
            y = y0 + j * Sy
            for k in rng:
                z = z0 + k * Sz
                dist[(i, j, k)] = min([f.distance(x, y, z) for f in itemlist])
    area = 0
    for i in rng0:
        for j in rng0:
            for k in rng0:
                area += sub_surface_area(
                    dist[i, j, k],
                    dist[i + 1, j, k],
                    dist[i + 1, j + 1, k],
                    dist[i, j + 1, k],
                    dist[i, j, k + 1],
                    dist[i + 1, j, k + 1],
                    dist[i + 1, j + 1, k + 1],
                    dist[i, j + 1, k + 1],
                    0,
                    Sx,
                    0,
                    Sy,
                    0,
                    Sz,
                )
    return area


def sub_surface_area(v0, v1, v2, v3, v4, v5, v6, v7, x0, x1, y0, y1, z0, z1):
    def allornone(lst):
        return all(lst) or not any(lst)

    # if all corners are inside -- return 0
    if allornone(
        [x <= options.ics_distance_threshold for x in [v0, v1, v2, v3, v4, v5, v6, v7]]
    ):
        return 0

    results = numpy.array([0.0] * 48)

    find_triangles = nrn_dll_sym("find_triangles")
    find_triangles.argtypes = [
        c_double,
        c_double,
        c_double,
        c_double,
        c_double,
        c_double,
        c_double,
        c_double,
        c_double,  # value0 - value7
        c_double,
        c_double,
        c_double,
        c_double,
        c_double,
        c_double,  # x0, x1, y0, y1, z0, z1
        POINTER(c_double),
    ]

    llgramarea = nrn_dll_sym("llgramarea")
    llgramarea.argtypes = [POINTER(c_double), POINTER(c_double), POINTER(c_double)]
    llgramarea.restype = c_double

    results_ptr = results.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

    count = find_triangles(
        abs(options.ics_distance_threshold),
        v0,
        v1,
        v2,
        v3,
        v4,
        v5,
        v6,
        v7,
        x0,
        x1,
        y0,
        y1,
        z0,
        z1,
        results_ptr,
    )

    pt0 = numpy.array([0.0] * 3)
    pt1 = numpy.array([0.0] * 3)
    pt2 = numpy.array([0.0] * 3)

    pt0_ptr = pt0.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
    pt1_ptr = pt1.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
    pt2_ptr = pt2.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

    results = results.reshape(16, 3)

    tri_i = 0
    area = 0
    for tri_i in range(count):
        pt0[:] = results[0 + 3 * tri_i]
        pt1[:] = results[1 + 3 * tri_i]
        pt2[:] = results[2 + 3 * tri_i]

        area += llgramarea(pt0_ptr, pt1_ptr, pt2_ptr) / 2.0

    return area



"""
Convert rst to a compressed dictionary suitable for NEURON + Python help system.
Run via:
>>> python3 parse_rst.py ./python/ help_data.dat
"""

import os
import sys
from pathlib import Path


class ParseRst(object):

    help_dictionary = {}

    def __init__(self, rst_path, out_file):
        self._rst_path = rst_path
        self._out_file = out_file
        self._filenames = Path(self._rst_path).glob("**/*.rst")

    @classmethod
    def handle_identifier(cls, lines, i, identifier):
        identifier = ".. %s::" % identifier
        line = lines[i]
        start = line.find(identifier)
        # print line, identifier, start
        # print identifier
        if start >= 0:
            name = line[start + len(identifier) :].strip()

            # print('%s -- %s' % (name, identifier))

            indent_line = lines[i + 1]
            while not indent_line.strip():
                i += 1
                indent_line = lines[i + 1]
            start = 0
            while start < len(indent_line):
                if indent_line[start] != " ":
                    break
                start += 1

            # TODO: store the body text
            body = []
            while i < len(lines) - 1:
                i += 1
                if lines[i].strip():
                    if lines[i][:start] == indent_line[:start]:
                        next_line = lines[i][start:]
                        if next_line[-1] == "\n":
                            next_line = next_line[:-1]
                        body.append(next_line)
                    else:
                        break
                else:
                    if not body or body[-1] != "\n":
                        body.append("\n")
            cls.help_dictionary[name] = "\n".join(body)

    def parse(self):
        for filename in self._filenames:
            with open(str(filename), encoding="utf-8") as f:
                lines = []
                for line in f:
                    if line[-1] == "\n":
                        line = line[:-1]
                    lines.append(line)
            i = 0
            while i < len(lines):
                for kind in ["method", "data", "class", "function"]:
                    ParseRst.handle_identifier(lines, i, kind)
                i += 1


if __name__ == "__main__":
    if len(sys.argv) == 3:
        rst_path = sys.argv[1]
        out_file = sys.argv[2]
    else:
        print("usage: python3 parse_rst.py <rst_folder_path> <output_file>")
        exit(1)

    try:
        ParseRst(rst_path, out_file).parse()
        with open(out_file, "wb") as f:
            import pickle
            import zlib

            compressed = zlib.compress(pickle.dumps(ParseRst.help_dictionary))
            f.write(compressed)
    except Exception:
        import traceback

        print(traceback.format_exc())
        exit(1)



import pytest


@pytest.fixture
def simple_model(neuron_instance):
    """A simple rxd model with species and regions and reactions."""

    h, rxd, data, save_path = neuron_instance
    dend = h.Section(name="dend")
    dend.diam = 2
    dend.nseg = 5
    dend.L = 5
    cyt = rxd.Region(dend, name="cyt", nrn_region="i")
    k = rxd.Species([cyt], name="k", d=1, charge=1, initial=140)
    paramA = rxd.Parameter([cyt], name="paramA", initial=1)
    paramB = rxd.Parameter([cyt], initial=0)
    decay = rxd.Rate(k, -0.1 * k)
    model = (dend, cyt, k, paramA, paramB, decay)
    yield (neuron_instance, model)


def test_reinit(simple_model):
    """Test rxd.re_init updates node values from NEURON values"""

    neuron_instance, model = simple_model
    h, rxd, data, save_path = neuron_instance
    dend, cyt, k, paramA, paramB, decay = model
    h.finitialize(-65)
    dend(0.5).ki = 0
    rxd.re_init()
    assert k[cyt].nodes(dend(0.5)).value[0] == 0


def test_reinit_cvode(simple_model):
    """Test rxd.re_init updates node values from NEURON values with CVode"""

    neuron_instance, model = simple_model
    h, rxd, data, save_path = neuron_instance
    dend, cyt, k, paramA, paramB, decay = model
    h.finitialize(-65)
    h.CVode().active(True)
    dend(0.5).ki = 0
    rxd.re_init()
    assert k[cyt].nodes(dend(0.5)).value[0] == 0


def test_reinit_3d(simple_model):
    """Test rxd.re_init updates node values from NEURON values in 3D"""

    neuron_instance, model = simple_model
    h, rxd, data, save_path = neuron_instance
    dend, cyt, k, paramA, paramB, decay = model
    rxd.set_solve_type(dimension=3)
    # check changing the units after initialization
    h.finitialize(-65)
    dend(0.5).ki = 0
    rxd.re_init()
    for nd in k[cyt].nodes(dend(0.5)):
        assert nd.value == 0


def test_reinit_3d_cvode(simple_model):
    """Test rxd.re_init updates node values from NEURON values in 3D with
    CVode"""

    neuron_instance, model = simple_model
    h, rxd, data, save_path = neuron_instance
    dend, cyt, k, paramA, paramB, decay = model
    rxd.set_solve_type(dimension=3)
    h.CVode().active(True)
    # check changing the units after initialization
    h.finitialize(-65)
    dend(0.5).ki = 0
    rxd.species._has_1d = False
    rxd.re_init()
    for nd in k[cyt].nodes(dend(0.5)):
        assert nd.value == 0



from neuron import h, gui
from neuron.expect_hocerr import expect_err
import sys

pc = h.ParallelContext()


def callback(arg):
    print("callback: arg={}".format(arg))
    if arg == 1:
        print(1 / 0)
    if arg == 2:
        print("calling sys.exit()")
        sys.exit(0)


def test_finithandler():
    soma = h.Section(name="soma")
    fih = h.FInitializeHandler(1, (callback, (0)))
    for i in range(2):
        h.finitialize()

    fih = h.FInitializeHandler(1, (callback, (1)))
    for i in range(1):
        expect_err("h.finitialize()")

    del fih  # a mysterious consequence of expect_err is that need this
    locals()  # in order for the callback to be deleted in time for next
    if __name__ == "__main__":
        print("another test")
        fih = h.FInitializeHandler(1, (callback, (2)))
        h.finitialize()
        print("this line gets printed in interactive mode")


def efun(v):
    if v.x[0] == 100.0:
        print(1 / 0)
    if v.x[0] == 200.0:
        print("calling sys.exit()")
        sys.exit(0)
    e = v.sumsq()
    return e


def test_praxis():
    vec = h.Vector([2, 3])
    h.attr_praxis(1e-5, 0.5, 0)
    print("x={} vec={}".format(efun(vec), vec.to_python()))
    x = h.fit_praxis(efun, vec)
    print("after fit_praxis x={} vec={}".format(x, vec.to_python()))

    vec.x[0] = 100.0
    expect_err("h.fit_praxis(efun, vec)")
    if __name__ == "__main__":
        vec.x[0] = 200.0
        print("another test")
        x = h.fit_praxis(efun, vec)
        print("this line gets printed in interactive mode")


def foo(i):
    i = int(i)
    if i == 2:
        1 / 0
    if i == 3:
        sys.exit(0)
    return list(range(i))


def test_py2n_component():
    h("objref po")
    h.po = foo
    h("po = po._(1)")
    assert type(h.po) == type([])
    h.po = foo
    a = h("po = po._(2)")
    assert a == 0
    if __name__ == "__main__":
        h.po = foo
        h("po = po._(3)")
        print("this test_py2n_component line gets printed in interactive mode")
    h("objref po")


def test_func_call():
    for sec in h.allsec():
        h.delete_section(sec=sec)
    soma = h.Section(name="soma")
    dend = h.Section(name="dend")
    axon = h.Section(name="axon")
    dend.connect(soma(0))
    dend.nseg = 5
    axon.connect(soma(1))
    axon.nseg = 3
    h.topology()
    invoke = False

    def f(x):
        y = x**2
        if invoke:
            if h.cas() == dend and x == 1.0:
                1 / 0
            if h.cas() == dend and x > 0.6:
                sys.exit(0)
        print("{}({})".format(h.cas(), x))
        return y

    # all points exist
    rvp = h.RangeVarPlot(f, dend(1.0), axon(1.0))
    vec = h.Vector()
    rvp.to_vector(vec)
    invoke = True

    # but no errors as those are marked non-existent
    h.RangeVarPlot(f, dend(1.0), axon(1.0))

    # now an error
    expect_err("rvp.to_vector(vec)")

    if __name__ == "__main__":
        invoke = False
        rvp = h.RangeVarPlot(f, dend(0.9), axon(1.0))
        invoke = True
        rvp.to_vector(vec)
        print("this test_func_call line gets printed in interactive mode")

    del soma, dend, axon, rvp, vec
    locals()


def f(x):
    y = x * x
    if x == 4:
        y = 1.0 / 0.0
    return x, y


def test_call_picklef():
    print("test_call_picklef")
    pc.runworker()
    for i in range(6):
        pc.submit(f, float(i))
    try:
        while pc.working():
            x, y = pc.pyret()
            print("{} = f({})".format(y, x))
    except Exception as e:
        print(e)


if __name__ == "__main__":
    test_praxis()
    test_finithandler()
    test_py2n_component()
    test_call_picklef()
    test_func_call()



import os
import re
import shutil
import subprocess
import sys
from collections import defaultdict
from distutils import log
from distutils.dir_util import copy_tree
from distutils.version import LooseVersion
from setuptools import Command, Extension
from setuptools import setup


class Components:
    RX3D = True
    IV = True
    MPI = True
    CORENRN = False  # still early support
    GPU = False  # still early support


if os.name != "posix":
    raise Exception(
        "Python NEURON distributions are currently only available "
        "for Mac and Linux systems (POSIX)"
    )

# Main source of the version. Dont rename, used by Cmake
try:
    # github actions somehow fails with check_output and python3

    # Official Versioning shall rely on annotated tags (don't use `--tags` or `--all`)
    # (please refer to NEURON SCM documentation)
    v = (
        subprocess.run(["git", "describe"], stdout=subprocess.PIPE)
        .stdout.strip()
        .decode()
    )

    __version__ = v[: v.rfind("-")].replace("-", ".") if "-" in v else v
    # allow to override version during development/testing
    if "NEURON_WHEEL_VERSION" in os.environ:
        __version__ = os.environ["NEURON_WHEEL_VERSION"]

except Exception as e:
    raise RuntimeError("Could not get version from Git repo : " + str(e))


# setup options must be checked for very early as it impacts imports
if "--disable-rx3d" in sys.argv:
    Components.RX3D = False
    sys.argv.remove("--disable-rx3d")

if "--disable-iv" in sys.argv:
    Components.IV = False
    sys.argv.remove("--disable-iv")

if "--disable-mpi" in sys.argv:
    Components.MPI = False
    sys.argv.remove("--disable-mpi")

if "--enable-coreneuron" in sys.argv:
    Components.CORENRN = True
    sys.argv.remove("--enable-coreneuron")

if "--enable-gpu" in sys.argv:
    Components.GPU = True
    sys.argv.remove("--enable-gpu")

if Components.RX3D:
    try:
        from Cython.Distutils import Extension as CyExtension
        from Cython.Distutils import build_ext
        import numpy
    except ImportError:
        log.error(
            "ERROR: RX3D wheel requires Cython and numpy. Please install beforehand"
        )
        sys.exit(1)
else:
    from setuptools.command.build_ext import build_ext


class CMakeAugmentedExtension(Extension):
    """
    Definition of an extension that depends on a project to be built using CMake
    Notice by default the cmake project is installed to build/cmake_install
    """

    def __init__(
        self,
        name,
        sources,
        cmake_proj_dir="",
        cmake_flags=None,
        cmake_collect_dirs=None,
        cmake_install_python_files="lib/python",
        **kw,
    ):
        """Creates a CMakeAugmentedExtension.

        Args:
            name: the full name of the extension module (e.g. neuron.hoc)
            sources: The cpp sources of the extension
            cmake_proj_dir: The location of the cmake project
            cmake_flags: A list of options to be passed to cmake
            cmake_collect_dirs: Upon cmake success, these dirs are copied
                as package data in a dir '.data'
            cmake_install_python_files: Path to be scanned for python
                modules to be added directly to the python package tree
        """
        Extension.__init__(self, name, sources, **kw)
        self.sourcedir = os.path.abspath(cmake_proj_dir)
        self.cmake_flags = cmake_flags or []
        self.cmake_install_prefix = os.path.abspath("build/cmake_install")
        self.cmake_collect_dirs = cmake_collect_dirs or []
        self.cmake_install_python_files = cmake_install_python_files
        self.cmake_done = False


class CMakeAugmentedBuilder(build_ext):
    """Builder which understands CMakeAugmentedExtension"""

    user_options = build_ext.user_options + [
        ("cmake-prefix=", None, "value for CMAKE_PREFIX_PATH"),
        ("cmake-defs=", None, "Additional CMake definitions, comma split"),
    ]
    boolean_options = build_ext.boolean_options + ["docs"]

    def initialize_options(self):
        build_ext.initialize_options(self)
        self.cmake_prefix = None
        self.cmake_defs = None
        self.docs = False

    def run(self, *args, **kw):
        """Execute the extension builder.
        In case the given extension is a CMakeAugmentedExtension launches
        the CMake building, sets the extension build environment and collects files.
        """
        for ext in self.extensions:

            if isinstance(ext, CMakeAugmentedExtension):
                if ext.cmake_done:
                    continue

                self._run_cmake(ext)  # Build cmake project

                if self.docs:
                    return

                # Collect project files to be installed
                # These go directly into final package, regardless of setuptools filters
                log.info("\n==> Collecting CMAKE files")
                rel_package = ext.name.split(".")[:-1]
                package_data_d = os.path.join(
                    self.build_lib, *(rel_package + [".data"])
                )

                # Python files created/configured by CMake
                src_py_dir = os.path.join(
                    ext.cmake_install_prefix, ext.cmake_install_python_files
                )
                if os.path.isdir(src_py_dir):
                    copy_tree(src_py_dir, self.build_lib)  # accepts existing dst dir
                    shutil.rmtree(src_py_dir)  # avoid being collected to data dir

                for d in ext.cmake_collect_dirs:
                    log.info("  - Collecting %s (and everything under it)", d)
                    src_dir = os.path.join(ext.cmake_install_prefix, d)
                    dst_dir = os.path.join(package_data_d, d)
                    if not os.path.isdir(dst_dir):
                        shutil.copytree(src_dir, dst_dir)
                log.info("==> Done building CMake project\n.")

                # Make the temp include paths in the building the extension
                ext.include_dirs += [
                    os.path.join(self.build_temp, inc_dir)
                    for inc_dir in ext.include_dirs
                    if not os.path.isabs(inc_dir)
                ]
                ext.cmake_done = True

        # Now build the extensions normally
        log.info("==> Building Python extensions")
        build_ext.run(self, *args, **kw)

    def _run_cmake(self, ext):
        cmake = self._find_cmake()
        cfg = "Debug" if self.debug else "Release"
        self.outdir = os.path.abspath(ext.cmake_install_prefix)

        log.info("Building lib to: %s", self.outdir)
        cmake_args = [
            # Generic options only. project options shall be passed as ext param
            "-DCMAKE_INSTALL_PREFIX=" + self.outdir,
            "-DPYTHON_EXECUTABLE=" + sys.executable,
            "-DCMAKE_BUILD_TYPE=" + cfg,
        ] + ext.cmake_flags
        # RTD neds quick config
        if self.docs and os.environ.get("READTHEDOCS"):
            cmake_args = ["-DNRN_ENABLE_MPI=OFF", "-DNRN_ENABLE_INTERVIEWS=OFF"]
        if self.docs:
            cmake_args.append("-DNRN_ENABLE_DOCS=ON")
            cmake_args.append("-DNRN_ENABLE_DOCS_WITH_EXTERNAL_INSTALLATION=ON")
        if self.cmake_prefix:
            cmake_args.append("-DCMAKE_PREFIX_PATH=" + self.cmake_prefix)
        if self.cmake_defs:
            cmake_args += ["-D" + opt for opt in self.cmake_defs.split(",")]

        build_args = ["--config", cfg, "--", "-j4"]  # , 'VERBOSE=1']

        env = os.environ.copy()
        env["CXXFLAGS"] = "{} -DVERSION_INFO='{}'".format(
            env.get("CXXFLAGS", ""), self.distribution.get_version()
        )
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        try:
            # Configure project
            subprocess.Popen("echo $CXX", shell=True, stdout=subprocess.PIPE)
            log.info("[CMAKE] cmd: %s", " ".join([cmake, ext.sourcedir] + cmake_args))
            subprocess.check_call(
                [cmake, ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env
            )
            if self.docs:
                # RTD will call sphinx for us. We just need notebooks and doxygen
                if os.environ.get("READTHEDOCS"):
                    subprocess.check_call(
                        ["make", "notebooks"], cwd=self.build_temp, env=env
                    )
                    subprocess.check_call(
                        ["make", "doxygen"], cwd=self.build_temp, env=env
                    )
                else:
                    subprocess.check_call(
                        ["make", "docs"], cwd=self.build_temp, env=env
                    )
            else:
                subprocess.check_call(
                    [cmake, "--build", ".", "--target", "install"] + build_args,
                    cwd=self.build_temp,
                    env=env,
                )
                subprocess.check_call(
                    [
                        ext.cmake_install_prefix + "/bin/neurondemo",
                        "-nopython",
                        "-nogui",
                        "-c",
                        "quit()",
                    ],
                    cwd=self.build_temp,
                    env=env,
                )
                # mac: libnrnmech of neurondemo need to point to relative libnrniv
                REL_RPATH = (
                    "@loader_path" if sys.platform[:6] == "darwin" else "$ORIGIN"
                )
                subprocess.check_call(
                    [
                        ext.sourcedir + "/packaging/python/fix_demo_libnrnmech.sh",
                        ext.cmake_install_prefix,
                        REL_RPATH,
                    ],
                    cwd=self.build_temp,
                    env=env,
                )
                if Components.GPU:
                    subprocess.check_call(
                        [
                            ext.sourcedir
                            + "/packaging/python/fix_target_processor_in_makefiles.sh",
                            ext.cmake_install_prefix,
                        ],
                        cwd=self.build_temp,
                        env=env,
                    )

        except subprocess.CalledProcessError as exc:
            log.error("Status : FAIL. Log:\n%s", exc.output)
            raise

    @staticmethod
    def _find_cmake():
        for candidate in ["cmake", "cmake3"]:
            try:
                out = subprocess.check_output([candidate, "--version"])
                cmake_version = LooseVersion(
                    re.search(r"version\s*([\d.]+)", out.decode()).group(1)
                )
                if cmake_version >= "3.15.0":
                    return candidate
            except OSError:
                pass

        raise RuntimeError("Project requires CMake >=3.15.0")


class Docs(Command):
    description = "Generate & optionally upload documentation to docs server"
    user_options = [("upload", None, "Upload to docs server")]
    finalize_options = lambda self: None

    def initialize_options(self):
        self.upload = False

    def run(self):
        # The extensions must be created inplace to inspect docs
        self.reinitialize_command("build_ext", inplace=1, docs=True)
        self.run_command("build_ext")
        if self.upload:
            self._upload()

    def _upload(self):
        pass


def setup_package():
    NRN_PY_ROOT = "share/lib/python"
    NRN_PY_SCRIPTS = os.path.join(NRN_PY_ROOT, "scripts")
    NRN_COLLECT_DIRS = ["bin", "lib", "include", "share"]

    docs_require = []  # sphinx, themes, etc
    maybe_rxd_reqs = ["numpy", "Cython"] if Components.RX3D else []
    maybe_docs = docs_require if "docs" in sys.argv else []
    maybe_test_runner = ["pytest-runner"] if "test" in sys.argv else []

    py_packages = [
        "neuron",
        "neuron.neuroml",
        "neuron.tests",
        "neuron.rxd",
        "neuron.crxd",
        "neuron.gui2",
    ] + (["neuron.rxd.geometry3d"] if Components.RX3D else [])

    REL_RPATH = "@loader_path" if sys.platform[:6] == "darwin" else "$ORIGIN"

    extension_common_params = defaultdict(
        list,
        library_dirs=["build/cmake_install/lib"],
        libraries=["nrniv", "nrnpython{}".format(sys.version_info[0])],
    )

    extensions = [
        CMakeAugmentedExtension(
            "neuron.hoc",
            ["src/nrnpython/inithoc.cpp"],
            cmake_collect_dirs=NRN_COLLECT_DIRS,
            cmake_flags=[
                "-DNRN_ENABLE_CORENEURON=" + ("ON" if Components.CORENRN else "OFF"),
                "-DNRN_ENABLE_INTERVIEWS=" + ("ON" if Components.IV else "OFF"),
                "-DIV_ENABLE_X11_DYNAMIC=" + ("ON" if Components.IV else "OFF"),
                "-DNRN_ENABLE_RX3D=OFF",  # Never build within CMake
                "-DNRN_ENABLE_MPI=" + ("ON" if Components.MPI else "OFF"),
                "-DNRN_ENABLE_MPI_DYNAMIC=" + ("ON" if Components.MPI else "OFF"),
                "-DNRN_ENABLE_PYTHON_DYNAMIC=ON",
                "-DNRN_ENABLE_MODULE_INSTALL=OFF",
                "-DNRN_ENABLE_REL_RPATH=ON",
                "-DLINK_AGAINST_PYTHON=OFF",
                "-DCMAKE_VERBOSE_MAKEFILE=OFF",
                "-DCORENRN_ENABLE_OPENMP=ON",  # TODO: manylinux portability questions
            ]
            + (
                [
                    "-DCORENRN_ENABLE_GPU=ON",
                    "-DCMAKE_C_COMPILER=nvc",  # use nvc and nvc++ for GPU support
                    "-DCMAKE_CXX_COMPILER=nvc++",
                    "-DCMAKE_CUDA_COMPILER=nvcc",
                ]
                if Components.GPU
                else []
            ),
            include_dirs=[
                "src",
                "src/oc",
                "src/nrnpython",
                "src/nrnmpi",
            ],
            extra_link_args=[
                # use relative rpath to .data/lib
                "-Wl,-rpath,{}".format(REL_RPATH + "/.data/lib/")
            ],
            **extension_common_params,
        )
    ]

    if Components.RX3D:
        include_dirs = ["share/lib/python/neuron/rxd/geometry3d", numpy.get_include()]

        # Cython files take a long time to compile with O2 so default O0
        # But pay the price if uploading distribution
        extra_compile_args = ["-O2" if "NRN_BUILD_FOR_UPLOAD" in os.environ else "-O0"]
        rxd_params = extension_common_params.copy()
        rxd_params["libraries"].append("rxdmath")
        rxd_params.update(
            dict(
                # Cython files take a long time to compile with O2 but this
                # is a distribution...
                extra_compile_args=extra_compile_args,
                extra_link_args=[
                    "-Wl,-rpath,{}".format(REL_RPATH + "/../../.data/lib/")
                ],
            )
        )

        log.info("RX3D compile flags %s" % str(rxd_params))

        extensions += [
            CyExtension(
                "neuron.rxd.geometry3d.graphicsPrimitives",
                ["share/lib/python/neuron/rxd/geometry3d/graphicsPrimitives.pyx"],
                **rxd_params,
            ),
            CyExtension(
                "neuron.rxd.geometry3d.ctng",
                ["share/lib/python/neuron/rxd/geometry3d/ctng.pyx"],
                include_dirs=include_dirs,
                **rxd_params,
            ),
            CyExtension(
                "neuron.rxd.geometry3d.surfaces",
                [
                    "share/lib/python/neuron/rxd/geometry3d/surfaces.pyx",
                    "src/nrnpython/rxd_marching_cubes.cpp",
                    "src/nrnpython/rxd_llgramarea.cpp",
                ],
                include_dirs=include_dirs,
                **rxd_params,
            ),
        ]

    log.info("RX3D is %s", "ENABLED" if Components.RX3D else "DISABLED")

    # package name
    package_name = "NEURON-gpu" if Components.GPU else "NEURON"

    # For CI, we want to build separate wheel with "-nightly" suffix
    package_name += os.environ.get("NEURON_NIGHTLY_TAG", "-nightly")

    # GPU wheels use patchelf to avoid duplicating NVIDIA runtime libraries when
    # using nrnivmodl.
    maybe_patchelf = ["patchelf"] if Components.GPU else []

    setup(
        name=package_name,
        version=__version__,
        package_dir={"": NRN_PY_ROOT},
        packages=py_packages,
        package_data={"neuron": ["*.dat"]},
        ext_modules=extensions,
        scripts=[
            os.path.join(NRN_PY_SCRIPTS, f)
            for f in os.listdir(NRN_PY_SCRIPTS)
            if f[0] != "_"
        ],
        cmdclass=dict(build_ext=CMakeAugmentedBuilder, docs=Docs),
        install_requires=["numpy>=1.9.3"] + maybe_patchelf,
        tests_require=["flake8", "pytest"],
        setup_requires=["wheel"] + maybe_docs + maybe_test_runner + maybe_rxd_reqs,
        dependency_links=[],
    )


def mac_osx_setenv():
    """Set MacOS environment to build high-compat wheels"""
    try:
        # Also ensure wheel package is avail
        import wheel
    except ImportError:
        from setuptools.dist import Distribution

        Distribution().fetch_build_eggs(["wheel"])
    from wheel.macosx_libfile import extract_macosx_min_system_version

    sdk_root = (
        subprocess.check_output(["xcrun", "--sdk", "macosx", "--show-sdk-path"])
        .decode()
        .strip()
    )
    log.info("Setting SDKROOT=%s", sdk_root)
    os.environ["SDKROOT"] = sdk_root

    # Match Python OSX framework
    py_osx_framework = extract_macosx_min_system_version(sys.executable)
    if py_osx_framework is None:
        py_osx_framework = [10, 9]
    if py_osx_framework[1] > 9:
        log.warn(
            "[ WARNING ] You are building a wheel with a Python built"
            " for a recent MACOS version (from brew?). Your wheel won't be portable."
            " Consider using an official Python build from python.org"
        )
    macos_target = "%d.%d" % tuple(py_osx_framework[:2])
    log.info("Setting MACOSX_DEPLOYMENT_TARGET=%s", macos_target)
    os.environ["MACOSX_DEPLOYMENT_TARGET"] = macos_target


if __name__ == "__main__":
    if sys.platform[:6] == "darwin":
        mac_osx_setenv()
    setup_package()



import weakref
from . import species, rxdmath, rxd, node, initializer
import numpy
import copy
from .generalizedReaction import (
    GeneralizedReaction,
    ref_list_with_mult,
    get_scheme_rate1_rate2_regions_custom_dynamics_mass_action,
)
from .rxdException import RxDException


class Reaction(GeneralizedReaction):
    def __init__(self, *args, **kwargs):
        """Specify a reaction to be added to the system.

        Examples:
            For 2 * H + O > H2O in a mass action reaction at rate k:
                r = rxd.Reaction(2 * H + O, H2O, k)

            To constrain the reaction to a specified list of regions,
            say to just the extracellular space (ext) and the cytosol (cyt),
            use the regions keyword, e.g.
                r = rxd.Reaction(2 * H + O, H2O, k, regions=[ext, cyt])

            For a bi-directional reaction, specify a backward reaction rate.
            e.g. if kf is the forward rate and kb is the backward rate, then:
                r = rxd.Reaction(2 * H + O, H2O, kf, kb)

            To use dynamics other than mass-action, add that mass_action=False
            flag and put the full formula instead of a mass-action rate for
            kf (and kb). E.g. Michaelis-Menten degradation
                r = rxd.Reaction(
                    dimer, decomposed, dimer / (k + diamer), mass_action=False
                )
        """

        # parse the arguments
        (
            scheme,
            rate1,
            rate2,
            regions,
            custom_dynamics,
            mass_action,
        ) = get_scheme_rate1_rate2_regions_custom_dynamics_mass_action(args, kwargs)

        # TODO: verify schemes use weakrefs
        self._scheme = scheme
        if custom_dynamics is not None and mass_action is not None:
            raise RxDException("Cannot specify both custom_dynamics and mass_action.")
        elif custom_dynamics is None and mass_action is None:
            custom_dynamics = False
        elif custom_dynamics is None and mass_action is not None:
            custom_dynamics = not mass_action
        self._custom_dynamics = custom_dynamics
        if scheme._dir == "<":
            rate_f, rate_b = 0, rate1
            if rate2 is not None:
                raise RxDException(
                    "unidirectional Reaction can have only one rate constant"
                )
        elif scheme._dir == "<>":
            rate_f, rate_b = rate1, rate2
            if rate2 is None:
                raise RxDException("bidirectional Reaction needs two rate constants")
        elif scheme._dir == ">":
            rate_f, rate_b = rate1, 0
            if rate2 is not None:
                raise RxDException(
                    "unidirectional Reaction can have only one rate constant"
                )
        else:
            raise RxDException("unknown reaction scheme direction: %r" % scheme._dir)

        self._original_rate_f = rate_f
        self._original_rate_b = rate_b
        for ar in [scheme, rate_f, rate_b]:
            try:
                if ar._voltage_dependent:
                    self._voltage_dependent = True
                    break
            except AttributeError:
                pass
        else:
            self._voltage_dependent = False
        self._membrane_flux = False
        self._dir = scheme._dir
        self._custom_dynamics = custom_dynamics

        self._trans_membrane = False
        if not hasattr(regions, "__len__"):
            regions = [regions]
        self._regions = regions
        rxd._register_reaction(self)

        # initialize self if the rest of rxd is already initialized
        if initializer.is_initialized():
            self._do_init()

    def _do_init(self):
        self._update_rates()
        self._update_indices()

    def _update_rates(self):
        lhs = self._scheme._lhs._items
        rhs = self._scheme._rhs._items
        if self._dir == "<":
            # TODO: remove this limitation (probably means doing with rate_b what done with rate_f and making sure _sources and _dests are correct
            raise RxDException(
                "pure reverse reaction currently not supported; reformulate as a forward reaction"
            )

        rate_f = rxdmath._ensure_arithmeticed(self._original_rate_f)
        rate_b = rxdmath._ensure_arithmeticed(self._original_rate_b)

        if not self._custom_dynamics:
            for k, v in zip(list(lhs.keys()), list(lhs.values())):
                if v == 1:
                    rate_f *= k
                else:
                    rate_f *= k**v
            if self._dir == "<>":
                for k, v in zip(list(rhs.keys()), list(rhs.values())):
                    if v == 1:
                        rate_b *= k
                    else:
                        rate_b *= k**v
        rate = rate_f - rate_b
        self._rate_arithmeticed = rate

        self._sources = ref_list_with_mult(lhs)
        self._dests = ref_list_with_mult(rhs)

        # Check to if it is an extracellular reaction
        from . import region, species

        # Was an ECS region was passed to to the constructor
        ecs_region = [r for r in self._regions if isinstance(r, region.Extracellular)]

        # Are any of of the sources or destinations passed to the constructor extracellular
        if not ecs_region:
            ecs_species = [
                s()
                for s in self._sources + self._dests
                if isinstance(s(), species.SpeciesOnExtracellular)
                or isinstance(s(), species._ExtracellularSpecies)
            ]
            if ecs_species:
                ecs_region = (
                    [ecs_species[0]._region]
                    if isinstance(ecs_species[0], species._ExtracellularSpecies)
                    else [ecs_species[0]._extracellular()._region]
                )

        # Are any of of the sources or destinations passed to the constructor defined on the ECS
        if not ecs_region:
            sps = [
                s()
                for s in self._sources + self._dests
                if isinstance(s(), species.Species)
            ]
            # only have an ecs reaction if all the species are defined on the ecs
            if sps and all(s._extracellular_instances for s in sps):
                # take an intersection of all the extracellular regions
                ecs_region = list(sps[0]._extracellular_instances.keys())
                for s in sps:
                    ecs_region = [
                        r for r in s._extracellular_instances.keys() if r in ecs_region
                    ]
        if ecs_region:
            self._rate_ecs, self._involved_species_ecs = rxdmath._compile(
                rate, ecs_region
            )

        # if a region is specified -- use it
        if self._regions and self._regions != [None]:
            self._react_regions = [
                r for r in self._regions if any(r._secs1d) or any(r._secs3d)
            ]
        else:
            # else find the common regions shared by all sources and  destinations
            self._react_regions = []
            regs = []
            for sptr in self._sources + self._dests:
                s = sptr() if isinstance(sptr(), species.Species) else sptr()._species()
                regs.append(
                    set([r for r in s._regions if any(r._secs1d) or any(r._secs3d)])
                )
            self._react_regions = list(set.intersection(*regs))
        if self._react_regions:
            self._rate, self._involved_species = rxdmath._compile(
                rate, self._react_regions
            )
        else:
            self._rate, self._involved_species = None, []

        # Species are in at most one region
        trans_membrane = (
            len(
                {
                    s()._region()
                    for s in self._sources + self._dests + self._involved_species
                    if isinstance(s(), species.SpeciesOnRegion)
                }
            )
            + len(
                {
                    s()._extracellular()._region
                    for s in self._sources + self._dests + self._involved_species
                    if isinstance(s(), species.SpeciesOnExtracellular)
                }
            )
            > 1
        )

        if trans_membrane:
            raise RxDException(
                "Reaction does not support multi-compartment dynamics. Use MultiCompartmentReaction."
            )

        # Recompile all the reactions in C
        if hasattr(self, "_mult"):
            rxd._compile_reactions()

    @property
    def f_rate(self):
        """Get or set the forward reaction rate"""
        return self._original_rate_f

    @property
    def b_rate(self):
        """Get or set the backward reaction rate"""
        return self._original_rate_b

    @f_rate.setter
    def f_rate(self, value):
        if self._dir not in ("<>", ">"):
            raise RxDException("no forward reaction in reaction scheme")
        self._original_rate_f = value
        self._update_rates()

    @b_rate.setter
    def b_rate(self, value):
        if self._dir not in ("<>", "<"):
            raise RxDException("no backward reaction in reaction scheme")
        self._original_rate_b = value
        self._update_rates()

    def __repr__(self):
        short_f = (
            self._original_rate_f._short_repr()
            if hasattr(self._original_rate_f, "_short_repr")
            else self._original_rate_f
        )
        short_b = (
            self._original_rate_b._short_repr()
            if hasattr(self._original_rate_b, "_short_repr")
            else self._original_rate_b
        )

        if len(self._regions) != 1 or self._regions[0] is not None:
            regions_short = (
                "[" + ", ".join(r._short_repr() for r in self._regions) + "]"
            )
            return "Reaction(%s, %s, rate_b=%s, regions=%s, custom_dynamics=%r)" % (
                self._scheme,
                short_f,
                short_b,
                regions_short,
                self._custom_dynamics,
            )
        else:
            return "Reaction(%s, %s, rate_b=%s, custom_dynamics=%r)" % (
                self._scheme,
                short_f,
                short_b,
                self._custom_dynamics,
            )

    def _do_memb_scales(self):
        # nothing to do since NEVER a membrane flux
        pass



import math
import numpy
import functools
import copy
import sys
from .rxdException import RxDException
from . import initializer


def _vectorized(f, objs):
    if hasattr(objs, "__len__"):
        return numpy.array([f(obj) for obj in objs])
    else:
        return f(objs)


def _vectorized2(f, objs1, objs2):
    if hasattr(objs1, "__len__"):
        return numpy.array([f(objA, objB) for objA, objB in zip(objs1, objs2)])
    else:
        return f(objs1, objs2)


def _erf(objs):
    return _vectorized(math.erf, objs)


def _erfc(objs):
    return _vectorized(math.erfc, objs)


def _factorial(objs):
    return _vectorized(math.factorial, objs)


def _gamma(objs):
    return _vectorized(math.gamma, objs)


def _lgamma(objs):
    return _vectorized(math.lgamma, objs)


def _power(objs1, objs2):
    # TODO? assumes numpy arrays; won't work for lists
    return objs1**objs2


def _neg(objs):
    return -objs


def analyze_reaction(r):
    if not isinstance(r, _Reaction):
        print(("%r is not a reaction" % r))
    else:
        print(("%r is a reaction:" % r))
        print(
            (
                "   lhs: ",
                ", ".join(
                    "%s[%d]" % (sp, c)
                    for sp, c in zip(
                        list(r._lhs._items.keys()), list(r._lhs._items.values())
                    )
                ),
            )
        )
        print(
            (
                "   rhs: ",
                ", ".join(
                    "%s[%d]" % (sp, c)
                    for sp, c in zip(
                        list(r._rhs._items.keys()), list(r._rhs._items.values())
                    )
                ),
            )
        )
        print(("   dir: ", r._dir))


# TODO: change this so that inputs are all automatically converted to numpy.array(s)
# _compile is called by the reaction (Reaction._update_rates)
# returns the rate and the species involved
def _compile(arith, region):
    initializer._do_init()
    # for extracellular reactions ensure the species are _ExtracellularSpecies
    arith = _ensure_arithmeticed(arith)
    # arith = arith._ensure_extracellular(extracellular,intracellular3d)
    s_by_reg = {}
    species_dict = {}
    for reg in region:
        # Check to see if region has both 1D and 3D
        if (
            hasattr(reg, "_secs1d")
            and reg._secs1d
            and hasattr(reg, "_secs3d")
            and reg._secs3d
        ):
            for instruction in ["do_1d", "do_3d"]:
                # TODO figure out what we are catching with attribute error
                try:
                    # If it is a hybrid model, we need to do semi compile for both the 1D and the 3D
                    # Checks to make sure the all species in arith are defined on the region
                    try:
                        s = arith._semi_compile(reg, instruction)
                    except KeyError:
                        continue
                    # s_by_reg[reg] = s
                    s_by_reg.setdefault(reg, [])
                    s_by_reg[reg].append(s)
                    arith._involved_species(species_dict)
                except AttributeError:
                    species_dict = {}
                    s = str(arith)
        else:
            if hasattr(reg, "_secs1d") and reg._secs1d:
                instruction = "do_1d"
            elif hasattr(reg, "_secs3d") and reg._secs3d:
                instruction = "do_3d"
            # Do extracellular
            else:
                instruction = None
            try:
                # Non-Hybrid model so there are no additional instructions and behavior is normal
                # Checks to make sure the all species in arith are defined on the region
                try:
                    s = arith._semi_compile(reg, instruction)
                except KeyError:
                    continue
                # s_by_reg[reg] = s
                s_by_reg.setdefault(reg, [])
                s_by_reg[reg].append(s)
                arith._involved_species(species_dict)
            except AttributeError:
                species_dict = {}
                s = str(arith)

    # C-version
    # Get the index rather than the key
    return (s_by_reg, list(species_dict.values()))
    # (functools.partial(eval(command), numpy, sys.modules[__name__]), species_dict.values())


def _ensure_arithmeticed(other):
    from . import species

    if isinstance(other, species._SpeciesMathable):
        other = _Arithmeticed(other)
    elif isinstance(other, _Reaction):
        raise RxDException("Cannot do arithmetic on a reaction")
    elif not isinstance(other, _Arithmeticed):
        other = _Arithmeticed(other, valid_reaction_term=False)
    return other


def _validate_reaction_terms(r1, r2):
    if not (r1._valid_reaction_term or r2._valid_reaction_term):
        raise RxDException("lhs=%r and rhs=%r not valid in a reaction" % (r1, r2))
    elif not r1._valid_reaction_term:
        raise RxDException("lhs=%r not valid in a reaction" % r1)
    elif not r2._valid_reaction_term:
        raise RxDException("rhs=%r not valid in a reaction" % r2)


class _Function:
    def __init__(self, obj, f, fname):
        self._obj = _ensure_arithmeticed(obj)
        self._f = f
        self._fname = fname

    def __repr__(self):
        return "%s(%r)" % (self._fname, self._obj)

    def _short_repr(self):
        try:
            return "%s(%s)" % (self._fname, self._obj._short_repr())
        except:
            return self.__repr__()

    def _semi_compile(self, region, instruction):
        return "%s(%s)" % (self._fname, self._obj._semi_compile(region, instruction))

    def _involved_species(self, the_dict):
        self._obj._involved_species(the_dict)

    def _ensure_extracellular(self, extracellular=None):
        if extracellular:
            from . import species

            item = self._obj
            if isinstance(item, species.Species):
                ecs_species = item[extracellular]._extracellular()
                items = ecs_species
            elif hasattr(item, "_ensure_extracellular"):
                item._ensure_extracellular(extracellular=extracellular)

    @property
    def _voltage_dependent(self):
        try:
            return self._obj._voltage_dependent
        except AttributeError:
            return False


class _Function2:
    def __init__(self, obj1, obj2, f, fname):
        self._obj1 = _ensure_arithmeticed(obj1)
        self._obj2 = _ensure_arithmeticed(obj2)
        self._f = f
        self._fname = fname

    def __repr__(self):
        return "%s(%r, %r)" % (self._fname, self._obj1, self._obj2)

    def _short_repr(self):
        try:
            return "%s(%s, %s)" % (
                self._fname,
                self._obj1._short_repr(),
                self._obj2._short_repr(),
            )
        except:
            return self.__repr__()

    def _semi_compile(self, region, instruction):
        return "%s(%s, %s)" % (
            self._fname,
            self._obj1._semi_compile(region, instruction),
            self._obj2._semi_compile(region, instruction),
        )

    def _involved_species(self, the_dict):
        self._obj1._involved_species(the_dict)
        self._obj2._involved_species(the_dict)

    def _ensure_extracellular(self, extracellular=None):
        if extracellular:
            from . import species

            for item in [self._obj1, self._obj2]:
                if isinstance(item, species.Species):
                    ecs_species = item[extracellular]._extracellular()
                    items = ecs_species
                elif hasattr(item, "_ensure_extracellular"):
                    item._ensure_extracellular(extracellular=extracellular)

    @property
    def _voltage_dependent(self):
        for item in [self._obj1, self._obj2]:
            try:
                if item._voltage_dependent:
                    return True
            except AttributeError:
                pass
        return False


# wrappers for the functions in module math from python 2.7
def acos(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.arccos", "acos"), valid_reaction_term=False
    )


def acosh(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.arccosh", "acosh"), valid_reaction_term=False
    )


def asin(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.arcsin", "asin"), valid_reaction_term=False
    )


def asinh(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.arcsinh", "asinh"), valid_reaction_term=False
    )


def atan(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.arctan", "atan"), valid_reaction_term=False
    )


def atan2(obj1, obj2):
    return _Arithmeticed(
        _Function2(obj1, obj2, "numpy.arctan2", "atan2"), valid_reaction_term=False
    )


def ceil(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.ceil", "ceil"), valid_reaction_term=False
    )


def copysign(obj1, obj2):
    return _Arithmeticed(
        _Function2(obj1, obj2, "numpy.copysign", "copysign"), valid_reaction_term=False
    )


def cos(obj):
    return _Arithmeticed(_Function(obj, "numpy.cos", "cos"), valid_reaction_term=False)


def cosh(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.cosh", "cosh"), valid_reaction_term=False
    )


def degrees(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.degrees", "degrees"), valid_reaction_term=False
    )


def erf(obj):
    return _Arithmeticed(
        _Function(obj, "rxdmath._erf", "erf"), valid_reaction_term=False
    )


def erfc(obj):
    return _Arithmeticed(
        _Function(obj, "rxdmath._erfc", "erfc"), valid_reaction_term=False
    )


def exp(obj):
    return _Arithmeticed(_Function(obj, "numpy.exp", "exp"), valid_reaction_term=False)


def expm1(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.expm1", "expm1"), valid_reaction_term=False
    )


def fabs(obj):
    return _Arithmeticed(_Function(obj, "abs", "fabs"), valid_reaction_term=False)


def factorial(obj):
    return _Arithmeticed(
        _Function(obj, "rxdmath._factorial", "factorial"), valid_reaction_term=False
    )


def floor(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.floor", "floor"), valid_reaction_term=False
    )


def fmod(obj1, obj2):
    return _Arithmeticed(
        _Function2(obj1, obj2, "numpy.fmod", "fmod"), valid_reaction_term=False
    )


def frexp(obj):
    raise RxDException("frexp not supported in this context")


def fsum(obj):
    raise RxDException("fsum not supported in this context")


def gamma(obj):
    return _Arithmeticed(
        _Function(obj, "rxdmath._gamma", "gamma"), valid_reaction_term=False
    )


def hypot(obj1, obj2):
    return _Arithmeticed(
        _Function2(obj1, obj2, "numpy.hypot", "hypot"), valid_reaction_term=False
    )


def isinf(obj):
    raise RxDException("isinf not supported in this context")


def isnan(obj):
    raise RxDException("isnan not supported in this context")


def ldexp(obj1, obj2):
    return _Arithmeticed(
        _Function2(obj1, obj2, "numpy.ldexp", "ldexp"), valid_reaction_term=False
    )


def lgamma(obj):
    return _Arithmeticed(
        _Function(obj, "rxdmath.lgamma", "lgamma"), valid_reaction_term=False
    )


def log(obj):
    return _Arithmeticed(_Function(obj, "numpy.log", "log"), valid_reaction_term=False)


def log10(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.log10", "log10"), valid_reaction_term=False
    )


def log1p(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.log1p", "log1p"), valid_reaction_term=False
    )


def modf(obj):
    raise RxDException("modf not supported in this context")


# this seems to be okay; just have to avoid using pow in any other context
def pow(obj1, obj2):
    return _Arithmeticed(
        _Function2(obj1, obj2, "rxdmath._power", "pow"), valid_reaction_term=False
    )


def radians(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.radians", "radians"), valid_reaction_term=False
    )


def sin(obj):
    return _Arithmeticed(_Function(obj, "numpy.sin", "sin"), valid_reaction_term=False)


def sinh(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.sinh", "sinh"), valid_reaction_term=False
    )


def sqrt(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.sqrt", "sqrt"), valid_reaction_term=False
    )


def tan(obj):
    return _Arithmeticed(_Function(obj, "numpy.tan", "tan"), valid_reaction_term=False)


def tanh(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.tanh", "tanh"), valid_reaction_term=False
    )


def trunc(obj):
    return _Arithmeticed(
        _Function(obj, "numpy.trunc", "trunc"), valid_reaction_term=False
    )


def vtrap(obj1, obj2):
    return _Arithmeticed(
        _Function2(obj1, obj2, "vtrap", "vtrap"), valid_reaction_term=False
    )


class _Product:
    def __init__(self, a, b):
        self._a = a
        self._b = b

    def __repr__(self):
        return "(%r)*(%r)" % (self._a, self._b)

    # Change any Species to _ExtracellularSpecies so _semi_compile gives the
    # _grid_id and not the species _id
    def _ensure_extracellular(self, extracellular=None, intracellular3d=None):
        p = _Product(self._a, self._b)
        if extracellular:
            from . import species

            # for item in [self._a, self._b]:
            # if isinstance(item,species.Species):
            #    ecs_species = item[extracellular]._extracellular()
            #    items = ecs_species
            # elif hasattr(item,'_ensure_extracellular'):
            # if hasattr(item,'_ensure_extracellular'):
            #    item._ensure_extracellular(extracellular=extracellular)
            if hasattr(self._a, "_ensure_extracellular"):
                p._a = self._a._ensure_extracellular(extracellular=extracellular)
            if hasattr(self._b, "_ensure_extracellular"):
                p._b = self._b._ensure_extracellular(extracellular=extracellular)

        """if intracellular3d:
            from . import species
            for item in [self._a, self._b]:
                if isinstance(item,species.Species):
                    ics_species = item._intracellular_instances[intracellular3d]
                    items = ics_species
                elif hasattr(item,'_ensure_extracellular'):
                    item._ensure_extracellular(intracellular3d=intracellular3d)"""
        if intracellular3d:
            from . import species

            if hasattr(self._a, "_ensure_extracellular"):
                p._a = self._a._ensure_extracellular(intracellular3d=intracellular3d)
            if hasattr(self._b, "_ensure_extracellular"):
                p._b = self._b._ensure_extracellular(intracellular3d=intracellular3d)
        return p

    @property
    def _voltage_dependent(self):
        for item in [self._a, self._b]:
            try:
                if item._voltage_dependent:
                    return True
            except AttributeError:
                pass
        return False

    def _semi_compile(self, region, instruction):
        return "(%s)*(%s)" % (
            self._a._semi_compile(region, instruction),
            self._b._semi_compile(region, instruction),
        )

    def _involved_species(self, the_dict):
        self._a._involved_species(the_dict)
        self._b._involved_species(the_dict)


class _Quotient:
    def __init__(self, a, b):
        self._a = a
        self._b = b

    def __repr__(self):
        return "(%r)/(%r)" % (self._a, self._b)

    # Change any Species to _ExtracellularSpecies so _semi_compile gives the
    # _grid_id and not the species _id
    def _ensure_extracellular(self, extracellular=None, intracellular3d=None):
        q = _Quotient(self._a, self._b)
        if extracellular:
            from . import species

            """for item in [self._a, self._b]:
                if isinstance(item,species.Species):
                    ecs_species = item[extracellular]._extracellular()
                    items = ecs_species
                elif hasattr(item,'_ensure_extracellular'):
                    item._ensure_extracellular(extracellular=extracellular)"""
            if hasattr(self._a, "_ensure_extracellular"):
                q._a = self._a._ensure_extracellular(extracellular=extracellular)
            if hasattr(self._b, "_ensure_extracellular"):
                q._b = self._b._ensure_extracellular(extracellular=extracellular)

        if intracellular3d:
            from . import species

            if hasattr(self._a, "_ensure_extracellular"):
                q._a = self._a._ensure_extracellular(intracellular3d=intracellular3d)
            if hasattr(self._b, "_ensure_extracellular"):
                q._b = self._b._ensure_extracellular(intracellular3d=intracellular3d)

        return q

    @property
    def _voltage_dependent(self):
        for item in [self._a, self._b]:
            try:
                if item._voltage_dependent:
                    return True
            except AttributeError:
                pass
        return False

    def _semi_compile(self, region, instruction):
        return "(%s)/(%s)" % (
            self._a._semi_compile(region, instruction),
            self._b._semi_compile(region, instruction),
        )

    def _involved_species(self, the_dict):
        self._a._involved_species(the_dict)
        self._b._involved_species(the_dict)


class _Reaction:
    def __init__(self, lhs, rhs, direction):
        self._lhs = lhs
        self._rhs = rhs
        self._dir = direction

    def __repr__(self):
        return "%s%s%s" % (str(self._lhs), self._dir, str(self._rhs))

    def __bool__(self):
        return False

    @property
    def _voltage_dependent(self):
        for item in [self._lhs, self._rhs]:
            try:
                if item._voltage_dependent:
                    return True
            except AttributeError:
                pass
        return False


class _Arithmeticed:
    def __init__(self, item, valid_reaction_term=True):
        if isinstance(item, dict):
            self._items = dict(item)
            self._original_items = dict(item)
        elif isinstance(item, _Reaction):
            raise RxDException("Cannot do arithmetic on a reaction")
        else:
            self._items = {item: 1}
            self._original_items = {item: 1}
        self._valid_reaction_term = valid_reaction_term
        self._compiled_form = None

    def _evaluate(self, location):
        if self._compiled_form is None:
            self._compiled_form = _compile(self)
        if len(location) != 3:
            raise RxDException(
                "_evaluate needs a (region, section, normalized position) triple"
            )
        region, sec, x = location
        concentrations = [
            numpy.array(sp()[region].nodes(sec)(x).concentration)
            for sp in self._compiled_form[1]
        ]
        value = self._compiled_form[0](*concentrations)
        if len(value) != 1:
            # this could happen in 3D
            raise RxDException("found %d values; expected 1." % len(value))
        return value[0]

    # Change any Species to _ExtracellularSpecies so _semi_compile gives the
    # _grid_id and not the species _id
    def _ensure_extracellular(self, extracellular=None, intracellular3d=None):
        new_arith = _Arithmeticed({})
        if extracellular and hasattr(self, "_items"):
            from . import species

            for item, count in zip(
                list(self._items.keys()), list(self._items.values())
            ):
                if count:
                    if isinstance(item, species.Species):
                        ecs_species = item[extracellular]._extracellular()
                        # self._items.pop(item)
                        new_arith._items[ecs_species] = count
                    elif hasattr(item, "_ensure_extracellular"):
                        new_arith._items[
                            item._ensure_extracellular(extracellular=extracellular)
                        ] = count
                    else:
                        new_arith._items[item] = count
        if intracellular3d and hasattr(self, "_items"):
            from . import species

            for item, count in zip(
                list(self._items.keys()), list(self._items.values())
            ):
                if count:
                    if isinstance(item, species.Species):
                        ics_species = item._intracellular_instances[intracellular3d]
                        # self._items.pop(item)
                        new_arith._items[ics_species] = count
                        # self._items[ics_species] = count
                    elif hasattr(item, "_ensure_extracellular"):
                        new_arith._items[
                            item._ensure_extracellular(intracellular3d=intracellular3d)
                        ] = count
                    else:
                        new_arith._items[item] = count
        return new_arith

    def _short_repr(self):
        from . import species

        items = []
        counts = []
        for item, count in zip(list(self._items.keys()), list(self._items.values())):
            if count:
                items.append(item)
                counts.append(count)
        result = ""
        for i, c in zip(items, counts):
            try:
                short_i = "%s" % i._short_repr()
            except:
                short_i = "%r" % i
            if result and c > 0:
                result += "+"
            if c == -1:
                result += "-(%s)" % short_i
            elif c != 1:
                result += "%d*(%s)" % (c, short_i)
            elif c == 1:
                result += short_i
        if not result:
            result = "0"
        return result

    def __repr__(self):
        from . import species

        items = []
        counts = []
        result = ""
        for item, count in zip(list(self._items.keys()), list(self._items.values())):
            if count:
                if isinstance(item, species._SpeciesMathable):
                    items.append(str(item))
                    counts.append(count)
                else:
                    items.append(repr(item))
                    counts.append(count)
        for i, c in zip(items, counts):
            if result and c > 0:
                result += "+"
            if c == -1:
                result += "-(%s)" % i
            elif c != 1:
                result += "%d*(%s)" % (c, i)
            elif c == 1:
                result += i
        if not result:
            result = "0"
        return result

    @property
    def _voltage_dependent(self):
        for item in self._items:
            try:
                if item._voltage_dependent:
                    return True
            except AttributeError:
                pass
        return False

    def _semi_compile(self, region, instruction):
        items = []
        counts = []
        items_append = items.append
        counts_append = counts.append
        for item, count in zip(list(self._items.keys()), list(self._items.values())):
            if count:
                try:
                    items_append(item._semi_compile(region, instruction))
                except AttributeError:
                    items_append("%r" % item)
                counts_append(count)
        result = ""
        for i, c in zip(items, counts):
            if result and c > 0:
                result += "+"
            if c == -1:
                result += "-(%s)" % i
            elif c != 1:
                result += "%d*(%s)" % (c, i)
            elif c == 1:
                result += i
        if not result:
            result = "0"
        return result

    def _involved_species(self, the_dict):
        for item, count in zip(list(self._items.keys()), list(self._items.values())):
            if count:
                try:
                    item._involved_species(the_dict)
                except AttributeError:
                    pass

    def _do_mul(self, other):
        if isinstance(other, int):
            items = dict(self._items)
            for i in items:
                items[i] *= other
            return _Arithmeticed(items, self._valid_reaction_term)
        else:
            other = _ensure_arithmeticed(other)
            return _Arithmeticed(_Product(self, other), False)

    def __mul__(self, other):
        return self._do_mul(other)

    def __rmul__(self, other):
        return self._do_mul(other)

    def __abs__(self):
        return _Arithmeticed(
            _Function(self, "numpy.abs", "fabs"), valid_reaction_term=False
        )

    def __pos__(self):
        return self

    def __neg__(self):
        return _Arithmeticed(
            _Function(self, "rxdmath._neg", "-"), valid_reaction_term=False
        )

    def __div__(self, other):
        other = _ensure_arithmeticed(other)
        return _Arithmeticed(_Quotient(self, other), False)

    def __rdiv__(self, other):
        other = _ensure_arithmeticed(other)
        return other / self

    def __truediv__(self, other):
        other = _ensure_arithmeticed(other)
        return _Arithmeticed(_Quotient(self, other), False)

    def __rtruediv__(self, other):
        other = _ensure_arithmeticed(other)
        return other / self

    def __pow__(self, other):
        return pow(self, other)

    def __ne__(self, other):
        other = _ensure_arithmeticed(other)
        _validate_reaction_terms(self, other)
        return _Reaction(self, other, "<>")

    def __gt__(self, other):
        other = _ensure_arithmeticed(other)
        _validate_reaction_terms(self, other)
        return _Reaction(self, other, ">")

    def __lt__(self, other):
        other = _ensure_arithmeticed(other)
        _validate_reaction_terms(self, other)
        return _Reaction(self, other, "<")

    def __add__(self, other):
        other = _ensure_arithmeticed(other)
        new_items = dict(self._items)
        for oitem in other._items:
            if oitem not in new_items:
                new_items[oitem] = other._items[oitem]
            else:
                new_items[oitem] += other._items[oitem]
        return _Arithmeticed(
            new_items, self._valid_reaction_term and other._valid_reaction_term
        )

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        other = _ensure_arithmeticed(other)
        new_items = dict(self._items)
        for oitem in other._items:
            if oitem not in new_items:
                new_items[oitem] = -other._items[oitem]
            else:
                new_items[oitem] -= other._items[oitem]
        return _Arithmeticed(new_items, False)

    def __rsub__(self, other):
        other = _ensure_arithmeticed(other)
        return other.__sub__(self)


class Vm(_Arithmeticed, object):
    """represent the membrane potential in rxd rates and reactions"""

    class _Vm(object):
        def __repr__(self):
            return "v"

        @property
        def _voltage_dependent(self):
            return True

    def __init__(self):
        super(Vm, self).__init__(Vm._Vm(), valid_reaction_term=True)


v = Vm()



#!/usr/bin/env python
"""
A generic wrapper to access nrn binaries from a python installation
Please create a softlink with the binary name to be called.
"""
import os
import shutil
import subprocess
import sys
from pkg_resources import working_set
from distutils.ccompiler import new_compiler
from sysconfig import get_config_vars, get_config_var


# This replaces the now depricated distutils.sysutils.customize_compiler
def _customize_compiler(compiler):
    """Do platform-sepcific customizations of compilers on unix platforms."""
    if compiler.compiler_type == "unix":
        (cc, cxx, cflags) = get_config_vars("CC", "CXX", "CFLAGS")
        if "CC" in os.environ:
            cc = os.environ["CC"]
        if "CXX" in os.environ:
            cxx = os.environ["CXX"]
        if "CFLAGS" in os.environ:
            cflags = cflags + " " + os.environ["CFLAGS"]
        cc_cmd = cc + " " + cflags
        # We update executables in compiler to take advantage of distutils arg splitting
        compiler.set_executables(compiler=cc_cmd, compiler_cxx=cxx)


def _set_default_compiler():
    """Set (dont overwrite) CC/CXX so that apps dont use the build-time ones"""
    ccompiler = new_compiler()
    _customize_compiler(ccompiler)
    # xcrun wrapper must bring all args
    if ccompiler.compiler[0] == "xcrun":
        ccompiler.compiler[0] = get_config_var("CC")
        ccompiler.compiler_cxx[0] = get_config_var("CXX")
    os.environ.setdefault("CC", ccompiler.compiler[0])
    os.environ.setdefault("CXX", ccompiler.compiler_cxx[0])


def _config_exe(exe_name):
    """Sets the environment to run the real executable (returned)"""

    package_name = "neuron"

    # determine package to find the install location
    if "neuron-gpu-nightly" in working_set.by_key:
        print("INFO : Using neuron-gpu-nightly Package (Alpha Developer Version)")
        package_name = "neuron-gpu-nightly"
    elif "neuron-gpu" in working_set.by_key:
        print("INFO : Using neuron-gpu Package (Alpha Version)")
        package_name = "neuron-gpu"
    elif "neuron-nightly" in working_set.by_key:
        print("INFO : Using neuron-nightly Package (Developer Version)")
        package_name = "neuron-nightly"
    elif "neuron" in working_set.by_key:
        package_name = "neuron"
    else:
        raise RuntimeError("NEURON package not found! Verify PYTHONPATH")

    NRN_PREFIX = os.path.join(
        working_set.by_key[package_name].location, "neuron", ".data"
    )
    os.environ["NEURONHOME"] = os.path.join(NRN_PREFIX, "share/nrn")
    os.environ["NRNHOME"] = NRN_PREFIX
    os.environ["CORENRNHOME"] = NRN_PREFIX
    os.environ["NRN_PYTHONEXE"] = sys.executable
    os.environ["CORENRN_PYTHONEXE"] = sys.executable
    os.environ["CORENRN_PERLEXE"] = shutil.which("perl")
    os.environ["NRNBIN"] = os.path.dirname(__file__)

    _set_default_compiler()
    return os.path.join(NRN_PREFIX, "bin", exe_name)


def _wrap_executable(output_name):
    """Create a wrapper for an executable in same dir. Requires renaming the original file.
    Executables are typically found under arch_name
    """
    release_dir = os.path.join(os.environ["NEURONHOME"], "demo/release")
    arch_name = next(os.walk(release_dir))[1][0]  # first dir
    file_path = os.path.join(arch_name, output_name)
    shutil.move(file_path, file_path + ".nrn")
    shutil.copy(__file__, file_path)


if __name__ == "__main__":
    exe = _config_exe(os.path.basename(sys.argv[0]))

    if exe.endswith("nrnivmodl"):
        # To create a wrapper for special (so it also gets ENV vars) we intercept nrnivmodl
        subprocess.check_call([exe, *sys.argv[1:]])
        _wrap_executable("special")
        sys.exit(0)

    if exe.endswith("special"):
        exe = os.path.join(
            sys.argv[0] + ".nrn"
        )  # original special is renamed special.nrn

    os.execv(exe, sys.argv)



from neuron import h

a = h.Section()
vec = h.Vector(10).indgen().add(100)
tp = []
for i in range(0, 10):
    tp.append(h.t1(a(0.5)))


a.v = 10

tp[2].f1()

# h.setpointer(a(.5)._ref_v, 'p1', tp[0])
tp[0]._ref_p1 = a(0.5)._ref_v
for i in range(1, 10):
    # h.setpointer(vec._ref_x[i], 'p1', tp[i])
    tp[i]._ref_p1 = vec._ref_x[i]

for i in range(0, 10):
    print(i, tp[i].p1, tp[i].f1())

tp[2].p1 = 25
print(vec[2], tp[2].p1)

z = h.t1()
import sys

try:
    h.setpointer(a(0.5)._ref_v, "p1", z)
except:
    print(sys.exc_info()[0], ": ", sys.exc_info()[1])

try:
    z._ref_p1 = a(0.5)._ref_v
except:
    print(sys.exc_info()[0], ": ", sys.exc_info()[1])



import pytest


def test_unicode_names(neuron_instance):
    """A test that species on a region that do not have a nrn_region can have
    an unicode name.
    """

    h, rxd, data, save_path = neuron_instance
    soma = h.Section(name="soma")
    cyt = rxd.Region(soma.wholetree(), name="cyt")
    α = rxd.Parameter(cyt, name="α", value=0.3)
    h.finitialize(-65)



from neuron import h

h.load_file("stdrun.hoc")


class Cell(object):
    def __init__(self):
        self.topology()
        self.subsets()
        self.geometry()
        self.biophys()
        self.synapses()

    def topology(self):
        self.soma = h.Section(cell=self)
        self.dend = h.Section(cell=self)
        self.dend.connect(self.soma)
        self.nseg = 5

    def subsets(self):
        self.all = h.SectionList()
        self.all.wholetree(sec=self.soma)

    def geometry(self):
        self.soma.L = 10
        self.soma.diam = 10
        self.dend.L = 500
        self.dend.diam = 1

    def biophys(self):
        for sec in self.all:
            sec.Ra = 100
            sec.cm = 1
        self.soma.insert("hh")

    def synapses(self):
        self.syn = h.ExpSyn(0.5, sec=self.dend)
        self.syn.e = 0
        self.syn.tau = 1

    def connectToTarget(self, syn):
        nc = h.NetCon(self.soma(0.5)._ref_v, syn, sec=self.soma)
        nc.threshold = -10
        return nc


pc = h.ParallelContext()
pc.set_gid2node(1, pc.id())
cell = Cell()
pc.cell(1, cell.connectToTarget(None))
stim = h.IClamp(cell.soma(0.5))
stim.delay = 0
stim.dur = 15
stim.amp = 0.1


from neuronmusic import *

out = publishEventOutput("out")
out.gid2index(1, 0)

pc.set_maxstep(1)
h.stdinit()
pc.psolve(20)



from neuron import h, crxd as rxd, gui
from neuron.crxd import v
from neuron.crxd.rxdmath import exp, log, vtrap
from math import pi

h.load_file("stdrun.hoc")

# parameters
h.celsius = 6.3
e = 1.60217662e-19
scale = 1e-14 / e
gnabar = 0.12 * scale  # molecules/um2 ms mV
gkbar = 0.036 * scale
gl = 0.0003 * scale
el = -54.3
q10 = 3.0 ** ((h.celsius - 6.3) / 10.0)

# sodium activation 'm'
alpha = 0.1 * vtrap(-(v + 40.0), 10)
beta = 4.0 * exp(-(v + 65) / 18.0)
mtau = 1.0 / (q10 * (alpha + beta))
minf = alpha / (alpha + beta)


# sodium inactivation 'h'
alpha = 0.07 * exp(-(v + 65.0) / 20.0)
beta = 1.0 / (exp(-(v + 35.0) / 10.0) + 1.0)
htau = 1.0 / (q10 * (alpha + beta))
hinf = alpha / (alpha + beta)

# potassium activation 'n'
alpha = 0.01 * vtrap(-(v + 55.0), 10.0)
beta = 0.125 * exp(-(v + 65.0) / 80.0)
ntau = 1.0 / (q10 * (alpha + beta))
ninf = alpha / (alpha + beta)

h.load_file("import3d.hoc")
cell = h.Import3d_SWC_read()
cell.input("c91662.swc")
i3d = h.Import3d_GUI(cell, 0)


class Cell:
    def move(self, offset=(0, 0, 0)):
        ox, oy, oz = offset[0], offset[1], offset[2]
        h.define_shape()
        xlo = ylo = zlo = xhi = yhi = zhi = None
        for sec in self.all:
            sec.nseg = 1
            n3d = sec.n3d()
            xs = [sec.x3d(i) + ox for i in range(n3d)]
            ys = [sec.y3d(i) + oy for i in range(n3d)]
            zs = [sec.z3d(i) + oz for i in range(n3d)]
            ds = [sec.diam3d(i) for i in range(n3d)]
            sec.pt3dclear()
            for x, y, z, d in zip(xs, ys, zs, ds):
                sec.pt3dadd(x, y, z, d)
            my_xlo, my_ylo, my_zlo = min(xs), min(ys), min(zs)
            my_xhi, my_yhi, my_zhi = max(xs), max(ys), max(zs)
            if xlo is None:
                xlo, ylo, zlo = my_xlo, my_ylo, my_zlo
                xhi, yhi, zhi = my_xhi, my_yhi, my_zhi
            else:
                xlo, ylo, zlo = min(xlo, my_xlo), min(ylo, my_ylo), min(zlo, my_zlo)
                xhi, yhi, zhi = max(xhi, my_xhi), max(yhi, my_yhi), max(zhi, my_zhi)
        return xlo, ylo, zlo, xhi, yhi, zhi


mycellA = Cell()
mycellB = Cell()
i3d.instantiate(mycellA)
i3d.instantiate(mycellB)

xloA, yloA, zloA, xhiA, yhiA, zhiA = mycellA.move((-100, 0, 0))
xloB, yloB, zloB, xhiB, yhiB, zhiB = mycellB.move((100, 0, 0))
xl, yl, zl = min(xloA, xloB) - 50, min(yloA, yloB) - 50, min(zloA, zloB) - 50
xh, yh, zh = max(xhiA, xhiB) + 50, max(yhiA, yhiB) + 50, max(zhiA, zhiB) + 50

# mod version
for sec in mycellB.all:
    sec.insert("hh")

# rxd version

# Where?
# intracellular
cyt = rxd.Region(h.allsec(), name="cyt", nrn_region="i")

# membrane
mem = rxd.Region(h.allsec(), name="cell_mem", geometry=rxd.membrane())

# extracellular
ecs = rxd.Extracellular(xl, yl, zl, xh, yh, zh, dx=50)

# Who?


def init(ics, ecs):
    return lambda nd: ecs if isinstance(nd, rxd.node.NodeExtracellular) else ics


# ions
k = rxd.Species([cyt, mem, ecs], name="k", d=1, charge=1, initial=init(54.4, 2.5))

na = rxd.Species([cyt, mem, ecs], name="na", d=1, charge=1, initial=init(10.0, 140.0))
x = rxd.Species([cyt, mem, ecs], name="x", charge=1)

ki, ko, nai, nao, xi, xo = k[cyt], k[ecs], na[cyt], na[ecs], x[cyt], x[ecs]

# gates
ngate = rxd.Species([cyt, mem], name="ngate", initial=0.24458654944007166)
mgate = rxd.Species([cyt, mem], name="mgate", initial=0.028905534475191907)
hgate = rxd.Species([cyt, mem], name="hgate", initial=0.7540796658225246)

# mycellA parameter
pA = rxd.Parameter(
    [cyt, mem],
    name="paramA",
    initial=lambda nd: 1 if nd.segment.sec in mycellA.all else 0,
)

# What
# gates
m_gate = rxd.Rate(mgate, (minf - mgate) / mtau)
h_gate = rxd.Rate(hgate, (hinf - hgate) / htau)
n_gate = rxd.Rate(ngate, (ninf - ngate) / ntau)

# Nernst potentials
ena = 1e3 * h.R * (h.celsius + 273.15) * log(nao / nai) / h.FARADAY  # 63
ek = 1e3 * h.R * (h.celsius + 273.15) * log(ko / ki) / h.FARADAY  # -74

gna = pA * gnabar * mgate**3 * hgate
gk = pA * gkbar * ngate**4

na_current = rxd.MultiCompartmentReaction(
    nai, nao, gna * (v - ena), mass_action=False, membrane=mem, membrane_flux=True
)
k_current = rxd.MultiCompartmentReaction(
    ki, ko, gk * (v - ek), mass_action=False, membrane=mem, membrane_flux=True
)
leak_current = rxd.MultiCompartmentReaction(
    xi, xo, pA * gl * (v - el), mass_action=False, membrane=mem, membrane_flux=True
)

# stimulate
stimA = h.IClamp(mycellA.soma[0](0.5))
stimA.delay = 50
stimA.amp = 5
stimA.dur = 50

stimB = h.IClamp(mycellB.soma[0](0.5))
stimB.delay = 50
stimB.amp = 5
stimB.dur = 50

# record
tvec = h.Vector().record(h._ref_t)

vvecA = h.Vector().record(mycellA.soma[0](0.5)._ref_v)
mvecA = h.Vector().record(mgate.nodes(mycellA.soma[0](0.5))[0]._ref_value)
nvecA = h.Vector().record(ngate.nodes(mycellA.soma[0](0.5))[0]._ref_value)
hvecA = h.Vector().record(hgate.nodes(mycellA.soma[0](0.5))[0]._ref_value)
kvecA = h.Vector().record(mycellA.soma[0](0.5)._ref_ik)
navecA = h.Vector().record(mycellA.soma[0](0.5)._ref_ina)

vvecB = h.Vector().record(mycellB.soma[0](0.5)._ref_v)
kvecB = h.Vector().record(mycellB.soma[0](0.5)._ref_ik)
navecB = h.Vector().record(mycellB.soma[0](0.5)._ref_ina)
mvecB = h.Vector().record(mycellB.soma[0](0.5).hh._ref_m)
nvecB = h.Vector().record(mycellB.soma[0](0.5).hh._ref_n)
hvecB = h.Vector().record(mycellB.soma[0](0.5).hh._ref_h)
tvec = h.Vector().record(h._ref_t)

# run
h.dt = 0.01
h.finitialize(-70)
# for i in range(1000):
#    h.fadvance()
#    print(h.t,mycellA.soma[0](0.5).v, mycellA.soma[0](0.5).ki, mycellA.soma[0](0.5).nai, mycellA.soma[0](0.5).xi)

h.continuerun(100)


if __name__ == "__main__":
    from matplotlib import pyplot

    # plot the results
    pyplot.ion()
    fig = pyplot.figure()
    pyplot.plot(tvec, vvecA, label="rxd")
    pyplot.plot(tvec, vvecB, label="mod")
    pyplot.legend()
    fig.set_dpi(200)

    fig = pyplot.figure()
    pyplot.plot(tvec, hvecA, "-b", label="h")
    pyplot.plot(tvec, mvecA, "-r", label="m")
    pyplot.plot(tvec, nvecA, "-g", label="n")
    pyplot.plot(tvec, hvecB, ":b")
    pyplot.plot(tvec, mvecB, ":r")
    pyplot.plot(tvec, nvecB, ":g")
    pyplot.legend()
    fig.set_dpi(200)

    fig = pyplot.figure()
    pyplot.plot(tvec, kvecA, "-b", label="k")
    pyplot.plot(tvec, navecA, "-r", label="na")
    pyplot.plot(tvec, kvecB, ":b")
    pyplot.plot(tvec, navecB, ":r")
    pyplot.legend()
    fig.set_dpi(200)

    fig = pyplot.figure()
    pyplot.plot(tvec, kvecA.as_numpy(), "-b", label="k")
    pyplot.plot(tvec, navecA.as_numpy(), "-r", label="na")
    pyplot.plot(tvec, kvecB, ":b")
    pyplot.plot(tvec, navecB, ":r")
    pyplot.legend()
    fig.set_dpi(200)

    fig = pyplot.figure()
    pyplot.plot(tvec, kvecA.as_numpy() - kvecB, "-b", label="k")
    pyplot.plot(tvec, navecA.as_numpy() - navecB, "-r", label="na")
    pyplot.legend()
    fig.set_dpi(200)



import pytest


def test_ecs_init(neuron_instance):
    """A test that reinitialization of extracellular species works when
    sections are added to the extracellular region after initialization.
    """
    h, rxd, data, save_path = neuron_instance

    ecs = rxd.Extracellular(-55, -55, -55, 55, 55, 55, dx=33)
    ca = rxd.Species(ecs, name="ca", charge=2)

    h.finitialize(-70)

    soma = h.Section(name="soma")
    h.finitialize(-70)

    assert True



import itertools
from threading import RLock

# TODO: monkey-patch everything that requires an init so only even attempts once


def _do_ion_register():
    from . import species

    for obj in species._all_species:
        obj = obj()
        if obj is not None:
            obj._ion_register()


_init_lock = RLock()
has_initialized = False


def _do_init():
    global has_initialized, _init_lock
    with _init_lock:
        if not has_initialized:
            from . import species, region, rxd

            if len(species._all_species) > 0:
                has_initialized = True
                # TODO: clean this up so not repetitive; can't do it super cleanly because of the multiple phases of species
                for obj in region._all_regions:
                    obj = obj()
                    if obj is not None:
                        obj._do_init()
                # NOTE: for species, must do all _do_init2s (1D alloc) before _do_init3s (3D alloc) to keep nodes arranged nicely
                for obj in species._all_species:
                    obj = obj()
                    if obj is not None:
                        obj._do_init1()
                for obj in species._all_species:
                    obj = obj()
                    if obj is not None:
                        obj._do_init2()
                for obj in species._all_species:
                    obj = obj()
                    if obj is not None:
                        obj._do_init3()
                for obj in species._all_species:
                    obj = obj()
                    if obj is not None:
                        obj._do_init4()
                for obj in species._all_species:
                    obj = obj()
                    if obj is not None:
                        obj._do_init5()
                for obj in rxd._all_reactions:
                    obj = obj()
                    if obj is not None:
                        obj._do_init()
                rxd._init()


def is_initialized():
    return has_initialized


def assert_initialized(msg=""):
    from . import species

    if not has_initialized and len(species._all_species) > 0:
        if len(msg):
            msg = ": " + msg
        from .rxdException import RxDException

        raise RxDException("invalid operation; rxd module not initialized" + msg)



from neuron import h, crxd as rxd
from math import pi
import time

h.load_file("stdrun.hoc")

h.CVode().active(True)
h.CVode().atol(1e-5)

rxd.options.enable.extracellular = True
# create cell1 where `x` will be created and leak out
cell1 = h.Section("cell1")
cell1.pt3dclear()
cell1.pt3dadd(-20, 0, 0, 10)
cell1.pt3dadd(-10, 0, 0, 10)
cell1.nseg = 11

# create cell2 where `x` will be pumped in and accumulate
cell2 = h.Section("cell2")
cell2.pt3dclear()
cell2.pt3dadd(10, 0, 0, 10)
cell2.pt3dadd(20, 0, 0, 10)
cell2.nseg = 11
# Where?
# the intracellular spaces
cyt = rxd.Region(
    h.allsec(),
    name="cyt",
    nrn_region="i",
    geometry=rxd.FractionalVolume(0.9, surface_fraction=1.0),
)

org = rxd.Region(h.allsec(), name="org", geometry=rxd.FractionalVolume(0.1))

cyt_org_membrane = rxd.Region(
    h.allsec(), name="mem", geometry=rxd.ScalableBorder(pi / 2.0, on_cell_surface=False)
)
cyt_ecs_membrane = rxd.Region(h.allsec(), name="cell_mem", geometry=rxd.membrane())

# the extracellular space
ecs = rxd.Extracellular(
    -55, -55, -55, 55, 55, 55, dx=10, volume_fraction=0.2, tortuosity=1.6
)

# Who?
x = rxd.Species(
    [cyt, org, cyt_org_membrane, cyt_ecs_membrane, ecs],
    name="x",
    d=1.0,
    charge=1,
    initial=0,
)
Xcyt = x[cyt]
Xorg = x[org]
Xecs = x[ecs]


# What? - produce X in cell 1
# parameter to limit production to cell 1
cell1_param = rxd.Parameter(
    org, initial=lambda node: 1.0 if node.segment.sec == cell1 else 0
)

# production with a rate following Michaels Menton kinetics
createX = rxd.Rate(Xorg, cell1_param[org] * 1.0 / (10.0 + Xorg))

# leak between organelles and cytosol
cyt_org_leak = rxd.MultiCompartmentReaction(
    Xcyt, Xorg, 1e4, 1e4, membrane=cyt_org_membrane, membrane_flux=False
)
from math import pi

e = 1.60217662e-19
scale = 1e-14 / e
cyt_ecs_pump1 = rxd.MultiCompartmentReaction(
    Xcyt,
    Xecs,
    0.1 * scale * Xcyt / (1.0 + Xcyt),
    -0.3 * scale * Xecs / (1.0 + Xecs),
    mass_action=False,
    membrane_flux=True,
    membrane=cyt_ecs_membrane,
)
cyt_ecs_pump2 = rxd.MultiCompartmentReaction(
    Xcyt,
    Xecs,
    1.4 * scale * Xcyt / (1.0 + Xcyt),
    0.8 * scale * Xecs / (1.0 + Xecs),
    mass_action=False,
    membrane_flux=True,
    membrane=cyt_ecs_membrane,
)
cyt_ecs_pump3 = rxd.MultiCompartmentReaction(
    Xcyt,
    Xecs,
    -0.5 * scale * Xcyt / (1.0 + Xcyt),
    0.5 * scale * Xecs / (1.0 + Xecs),
    mass_action=False,
    membrane_flux=True,
    membrane=cyt_ecs_membrane,
)


# record the concentrations in the cells
t_vec = h.Vector()
t_vec.record(h._ref_t)
cell1_X = h.Vector()
cell1_X.record(cell1(0.5)._ref_xi)
cell1_Xorg = h.Vector()
cell1_Xorg.record(Xorg.nodes(cell1)(0.5)[0]._ref_concentration)
cell1V = h.Vector()
cell1V.record(cell1(0.5)._ref_v)


cell2_X = h.Vector()
cell2_X.record(cell2(0.5)._ref_xi)
cell2_Xorg = h.Vector()
cell2_Xorg.record(Xorg.nodes(cell2)(0.5)[0]._ref_concentration)
cell2V = h.Vector()
cell2V.record(cell2(0.5)._ref_v)


# run and plot the results
h.finitialize(-65)
h.continuerun(1000)
if __name__ == "__main__":
    from matplotlib import pyplot

    fig = pyplot.figure()
    pyplot.subplot(2, 2, 1)
    pyplot.plot(t_vec, cell1_X, label="cyt")
    pyplot.plot(t_vec, cell1_Xorg, label="org")
    pyplot.legend()
    pyplot.xlabel("t (ms)")
    pyplot.ylabel("cell 1 x (mM)")

    pyplot.subplot(2, 2, 2)
    pyplot.plot(t_vec, cell2_X, label="cyt")
    pyplot.plot(t_vec, cell2_Xorg, label="org")
    pyplot.xlabel("t (ms)")
    pyplot.ylabel("cell 2 x (mM)")

    pyplot.subplot(2, 2, 3)
    pyplot.plot(t_vec, cell1V, label="cell1")
    pyplot.xlabel("t (ms)")
    pyplot.ylabel("cell 1 V$_m$ (mV)")

    pyplot.subplot(2, 2, 4)
    pyplot.plot(t_vec, cell2V, label="cell2")
    pyplot.xlabel("t (ms)")
    pyplot.ylabel("cell 2 V$_m$ (mV)")

    # pyplot.subplot(3, 1, 2)
    # pyplot.imshow(x[ecs].states3d.mean(2).T, extent=x[ecs].extent('xy'), origin="lower", aspect='equal')
    fig.tight_layout()
    pyplot.show()
